package org.processmining.logenhancement.abstraction;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension.StandardModel;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.log.utils.XUtils;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.utils.GraphIterator;
import org.processmining.models.graphbased.directed.utils.GraphIterator.EdgeAcceptor;
import org.processmining.models.graphbased.directed.utils.GraphIterator.NodeAcceptor;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.xesalignmentextension.XAlignmentExtension.MoveType;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignedLog;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignment;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignmentMove;
import org.processmining.xesalignmentextension.XDataAlignmentExtension;

import com.google.common.collect.BiMap;
import com.google.common.collect.ImmutableSet;

public class LogAbstracterBasedOnAlignment {

	public final static class ActivityInstance {

		private final Pattern pattern;
		private final long patternInstance;

		private long startTime = Long.MAX_VALUE;
		private long endTime = 0l;
		private Map<String, XAttribute> attributes = new HashMap<>();
		private List<XAlignmentMove> moves = new ArrayList<XAlignmentMove>();

		private ActivityInstance(Pattern pattern, long patternInstance) {
			super();
			this.pattern = pattern;
			this.patternInstance = patternInstance;
		}

		public String getName() {
			return pattern.getName();
		}

		public String getInstance() {
			return String.valueOf(patternInstance);
		}

		public long getStartTime() {
			return startTime;
		}

		public long getEndTime() {
			return endTime;
		}

		public Map<String, XAttribute> getAttributes() {
			return attributes;
		}

		public List<XAlignmentMove> getMoves() {
			return moves;
		}

		public void addMove(XAlignmentMove move) {
			if (move.getType() == MoveType.SYNCHRONOUS) {
				updateTimestamps(move);
				storeAttributes(move);
			}
			moves.add(move);
		}

		private void storeAttributes(XAlignmentMove move) {
			for (XAttribute attribute : move.getEvent().getAttributes().values()) {
				if (!XUtils.isStandardExtensionAttribute(attribute) && !isAlignmentExtensionAttribute(attribute)) {
					Boolean isModelAttribute = isModelAttribute(attribute);
					if (isModelAttribute) {
						attributes.put(attribute.getKey(), attribute);
					}
				}
			}
		}

		private void updateTimestamps(XAlignmentMove move) {
			Date timestamp = XUtils.getTimestamp(move.getEvent());
			if (timestamp != null) {
				startTime = Math.min(timestamp.getTime(), startTime);
				endTime = Math.max(timestamp.getTime(), endTime);
			}
		}

		public String toString() {
			return getName() + "-" + getInstance();
		}

	}

	public final static class Pattern implements Iterable<Transition> {

		private final String name;
		private final Set<Transition> transitionsInPattern = new HashSet<>();
		private final Transition startTransition;
		private final Transition endTransition;

		public Pattern(String name, Transition startTransition, Transition endTransition) {
			this.name = name;
			this.startTransition = startTransition;
			this.endTransition = endTransition;
		}

		public void addTransition(Transition t) {
			transitionsInPattern.add(t);
		}

		public Iterator<Transition> iterator() {
			return transitionsInPattern.iterator();
		}

		public String getName() {
			return name;
		}

		public Transition getStartTransition() {
			return startTransition;
		}

		public Transition getEndTransition() {
			return endTransition;
		}

		//identity equals and hashCode is sufficient!

		public String toString() {
			return getName();
		}

	}


	protected final Set<String> mappedEventClassIds;
	protected final Map<String, Transition> activityIdToTransition = new HashMap<>();

	protected final Set<Pattern> patternSet;

	protected final Map<String, Pattern> activityIdToPattern = new HashMap<>();
	protected final Map<String, Pattern> startIdToPattern = new HashMap<>();
	protected final Map<String, Pattern> endIdToPattern = new HashMap<>();

	protected final Map<Pattern, Deque<ActivityInstance>> instances = new HashMap<LogAbstracterBasedOnAlignment.Pattern, Deque<ActivityInstance>>();

	private boolean keepUnmappedEvents = false;
	private double errorRateLimit = 1.0;
	private XFactory factory = XFactoryRegistry.instance().currentDefault();

	public LogAbstracterBasedOnAlignment(final BiMap<Transition, String> startPointsMap,
			final BiMap<Transition, String> endPointsMap, TransEvClassMapping transitionMapping) throws PatternStructureException {
		super();
		this.mappedEventClassIds = createMappedClasses(transitionMapping);		
		this.patternSet = collectTransitionsInPatterns(startPointsMap, endPointsMap);

		for (Pattern pattern : patternSet) {
			for (Transition t : pattern) {
				activityIdToPattern.put(t.getLocalID().toString(), pattern);
				activityIdToTransition.put(t.getLocalID().toString(), t);
			}
		}

		for (Pattern pattern : patternSet) {
			startIdToPattern.put(pattern.getStartTransition().getLocalID().toString(), pattern);
		}

		for (Pattern pattern : patternSet) {
			endIdToPattern.put(pattern.getEndTransition().getLocalID().toString(), pattern);
		}
	}

	private static Set<String> createMappedClasses(TransEvClassMapping transitionMapping) {
		Set<String> eventClassIds = new HashSet<>();
		for (XEventClass eventClass: transitionMapping.values()) {
			eventClassIds.add(eventClass.getId());
		}
		return ImmutableSet.copyOf(eventClassIds);
	}

	private static Set<Pattern> collectTransitionsInPatterns(final BiMap<Transition, String> startPointsMap,
			final BiMap<Transition, String> endPointsMap) throws PatternStructureException {
		final Set<Pattern> patternSet = new HashSet<>();
		for (Entry<Transition, String> entry : startPointsMap.entrySet()) {
			Transition startTransition = entry.getKey();
			Transition endTransition = endPointsMap.inverse().get(entry.getValue());
			if (endTransition != null) {
				Pattern pattern = new Pattern(entry.getValue(), startTransition, endTransition);
				patternSet.add(pattern);
				Deque<Transition> currentTransitions = new ArrayDeque<Transition>();
				currentTransitions.push(startTransition);
				Set<Transition> seen = new HashSet<Transition>();
				while (!currentTransitions.isEmpty()) {
					Transition transition = currentTransitions.pop();
					pattern.addTransition(transition);
					if (transition != endTransition) {
						Collection<PetrinetNode> successors = getSuccessors(transition);
						boolean reachedEnd = successors.size() == 1
								&& successors.iterator().next().equals(endTransition);
						if (!reachedEnd) {
							for (PetrinetNode node : successors) {
								if (!seen.contains(node)) {
									seen.add((Transition) node);
									currentTransitions.push((Transition) node);
								}
							}
						}
					}
				}
				pattern.addTransition(endTransition);
			} else {
				throw new PatternStructureException("Missing endpoint for pattern " + entry.getValue());
			}
		}
		return patternSet;
	}

	private static Collection<PetrinetNode> getSuccessors(Transition transition) {

		final NodeAcceptor<PetrinetNode> nodeAcceptor = new NodeAcceptor<PetrinetNode>() {
			public boolean acceptNode(PetrinetNode node, int depth) {
				return (depth != 0) && (node instanceof Transition);
			}
		};

		return GraphIterator.getDepthFirstSuccessors(transition, transition.getGraph(),
				new EdgeAcceptor<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>>() {

					public boolean acceptEdge(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge,
							int depth) {
						return !nodeAcceptor.acceptNode(edge.getSource(), depth);
					}
				}, nodeAcceptor);
	}

	public XLog abstractBasedOnPatterns(XAlignedLog alignedLog) throws PatternStructureException {

		XLog abstractedLog = createInitialAbstractedLog(alignedLog);

		long patternInstance = 0;

		for (XAlignment alignment : alignedLog) {

			XTrace trace = getFactory().createTrace();
			copyNonAlignmentAttributes(alignment.getTrace(), trace);
			abstractedLog.add(trace);

			for (XAlignmentMove move : alignment) {

				if (move.getType() != MoveType.LOG) {

					// Get pattern
					Pattern pattern = activityIdToPattern.get(move.getActivityId());

					if (pattern != null) {

						// Check for patterns that started
						if (startIdToPattern.containsKey(move.getActivityId())) {
							pushInstance(instances, pattern, patternInstance++);
						}

						// Add move to correct activity instance
						ActivityInstance instance = peekInstance(instances, pattern);
						if (instance != null) {
							instance.addMove(move);
						} else {
							throw new PatternStructureException(
									"Missing start transition, expected activity instance for " + pattern.getName()
											+ "! For move " + move + " in Trace " + alignment);
						}

						// Check for patterns that are finished

						if (endIdToPattern.containsKey(move.getActivityId())) {

							// Remove current activity instance
							instance = popInstance(instances, pattern);

							if (isWithinErrorRateLimit(instance)) {

								// Abstract event
								abstractPattern(trace, instance);

							} else if (keepUnmappedEvents) {
								for (XAlignmentMove recordedMove : instance.getMoves()) {
									if (recordedMove.getType() == MoveType.SYNCHRONOUS) { // sync moves are observable
										trace.insertOrdered(getOriginalEvent(recordedMove));
									}
								}
							}
						}

					} else if (move.getType() == MoveType.SYNCHRONOUS) {
						throw new PatternStructureException("Unkown model/synchronous move " + move);
					}

				} else if (keepUnmappedEvents) {
					if (!mappedEventClassIds.contains(move.getEventClassId())) {
						trace.add(getOriginalEvent(move));	
					}					
				}
			}

			for (Deque<ActivityInstance> queue : instances.values()) {
				if (!queue.isEmpty()) {
					throw new RuntimeException("Unfinished activity instances leftover " + queue.toString());
				}
			}

		}
		return abstractedLog;
	}

	protected static ActivityInstance popInstance(Map<Pattern, Deque<ActivityInstance>> instances, Pattern pattern) {
		Deque<ActivityInstance> deque = instances.get(pattern);
		if (deque != null) {
			return deque.pop();
		} else {
			return null;
		}
	}

	protected static ActivityInstance peekInstance(Map<Pattern, Deque<ActivityInstance>> instances, Pattern pattern) {
		Deque<ActivityInstance> deque = instances.get(pattern);
		if (deque != null) {
			return deque.peek();
		} else {
			return null;
		}
	}

	protected static ActivityInstance pushInstance(Map<Pattern, Deque<ActivityInstance>> instances, Pattern pattern,
			long patternInstance) {
		ActivityInstance activityInstance = new ActivityInstance(pattern, patternInstance);
		Deque<ActivityInstance> deque = instances.get(pattern);
		if (deque == null) {
			deque = new ArrayDeque<LogAbstracterBasedOnAlignment.ActivityInstance>();
			instances.put(pattern, deque);
		}
		deque.push(activityInstance);
		return activityInstance;
	}

	protected XLog createInitialAbstractedLog(XAlignedLog alignedLog) {
		XLog abstractedLog = factory.createLog();
		copyNonAlignmentAttributes(alignedLog.getLog(), abstractedLog);

		abstractedLog.getExtensions().add(XLifecycleExtension.instance());
		abstractedLog.getExtensions().add(XTimeExtension.instance());
		abstractedLog.getExtensions().add(XConceptExtension.instance());

		abstractedLog.getGlobalEventAttributes().add(XConceptExtension.ATTR_NAME);
		abstractedLog.getGlobalEventAttributes().add(XConceptExtension.ATTR_INSTANCE);
		abstractedLog.getGlobalEventAttributes().add(XLifecycleExtension.ATTR_TRANSITION);
		abstractedLog.getGlobalEventAttributes().add(XTimeExtension.ATTR_TIMESTAMP);

		return abstractedLog;
	}

	protected void abstractPattern(XTrace trace, ActivityInstance instance) {
		XEvent startEvent = factory.createEvent();
		XConceptExtension.instance().assignName(startEvent, instance.getName());
		XConceptExtension.instance().assignInstance(startEvent, instance.getInstance());
		XLifecycleExtension.instance().assignStandardTransition(startEvent, StandardModel.START);

		for (XAttribute attribute : instance.getAttributes().values()) {
			if (!startEvent.getAttributes().containsKey(attribute.getKey())) {
				clearMeta(attribute);
				startEvent.getAttributes().put(attribute.getKey(), attribute);
			}
		}

		// Start
		if (instance.getStartTime() != Long.MAX_VALUE) {
			XTimeExtension.instance().assignTimestamp(startEvent, instance.getStartTime());
			trace.insertOrdered(startEvent);
		} else {
			// No start timestamp available just add after current event
			trace.add(startEvent);
		}

		XEvent completeEvent = factory.createEvent();
		XConceptExtension.instance().assignName(completeEvent, instance.getName());
		XConceptExtension.instance().assignInstance(completeEvent, instance.getInstance());
		XLifecycleExtension.instance().assignStandardTransition(completeEvent, StandardModel.COMPLETE);

		for (XAttribute attribute : instance.getAttributes().values()) {
			if (!completeEvent.getAttributes().containsKey(attribute.getKey())) {
				clearMeta(attribute);
				completeEvent.getAttributes().put(attribute.getKey(), attribute);
			}
		}

		// Complete
		if (instance.getEndTime() != 0) {
			XTimeExtension.instance().assignTimestamp(completeEvent, instance.getEndTime());
			trace.insertOrdered(completeEvent);
		} else {
			// No complete timestamp available just add after current event
			trace.add(completeEvent);
		}
	}

	protected boolean isWithinErrorRateLimit(ActivityInstance instance) {
		double errorRate = computeErrorRate(instance);
		return Double.compare(errorRate, errorRateLimit) < 0;
	}

	protected static double computeErrorRate(ActivityInstance instance) {
		int numMoves = 0;
		int numModelMoves = 0;
		for (XAlignmentMove move : instance.getMoves()) {
			if (move.isObservable()) {
				numMoves++;
				if (move.getType() == MoveType.MODEL) {
					numModelMoves++;
				}
			}
		}
		return numModelMoves / (double) numMoves;
	}

	protected static void clearMeta(XAttribute attribute) {
		if (attribute.hasAttributes()) {
			Iterator<XAttribute> iterator = attribute.getAttributes().values().iterator();
			while (iterator.hasNext()) {
				XAttribute metaAttribute = iterator.next();
				if (metaAttribute.getKey().startsWith("dataalignment")) {
					iterator.remove();
				}

			}
		}
	}

	protected static boolean isModelAttribute(XAttribute attribute) {
		Boolean isModelAttribute = XDataAlignmentExtension.instance().extractIsModelAttribute(attribute);
		if (isModelAttribute != null) {
			return isModelAttribute;
		}
		return false;
	}

	protected static void copyNonAlignmentAttributes(XAttributable source, XAttributable target) {
		for (XAttribute a : source.getAttributes().values()) {
			if (!isAlignmentExtensionAttribute(a)) {
				XUtils.putAttribute(target, (XAttribute) a.clone());
			}
		}
	}

	protected XEvent getOriginalEvent(XAlignmentMove move) {
		XEvent event = factory.createEvent();
		XEvent alignedEvent = move.getEvent();

		for (XAttribute attr : alignedEvent.getAttributes().values()) {
			if (!isAlignmentExtensionAttribute(attr)) {
				event.getAttributes().put(attr.getKey(), attr);
			}
		}

		XConceptExtension.instance().assignName(event, move.getEventClassId());

		return event;
	}

	protected static boolean isAlignmentExtensionAttribute(XAttribute a) {
		return a.getKey().startsWith("alignment:") || a.getKey().startsWith("dataalignment");
	}

	public double getErrorRateLimit() {
		return errorRateLimit;
	}

	public void setErrorRateLimit(double errorRateLimit) {
		this.errorRateLimit = errorRateLimit;
	}

	public boolean isKeepUnmappedEvents() {
		return keepUnmappedEvents;
	}

	public void setKeepUnmappedEvents(boolean keepUnmappedEvents) {
		this.keepUnmappedEvents = keepUnmappedEvents;
	}

	public XFactory getFactory() {
		return factory;
	}

	public void setFactory(XFactory factory) {
		this.factory = factory;
	}

}
