package org.processmining.logenhancement.abstraction;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;

import javax.lang.model.SourceVersion;
import javax.swing.JLabel;
import javax.swing.JOptionPane;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.logenhancement.abstraction.editor.AbstractionModelBuilder;
import org.processmining.logenhancement.abstraction.model.AbstractionModel;
import org.processmining.logenhancement.abstraction.model.AbstractionPattern;
import org.processmining.logenhancement.abstraction.model.PatternCompiler;
import org.processmining.logenhancement.abstraction.model.PatternSimplifier;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParser;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRoot;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.VariableAccess;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.DataConformance.framework.VariableMatchCosts;
import org.processmining.plugins.balancedconformance.BalancedDataXAlignmentPlugin;
import org.processmining.plugins.balancedconformance.config.BalancedProcessorConfiguration;
import org.processmining.plugins.balancedconformance.controlflow.ControlFlowAlignmentException;
import org.processmining.plugins.balancedconformance.dataflow.exception.DataAlignmentException;
import org.processmining.plugins.balancedconformance.observer.DataConformancePlusObserverImpl;
import org.processmining.plugins.balancedconformance.ui.DataConformanceConfigUIUtils;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;
import org.processmining.plugins.utils.ProvidedObjectHelper;
import org.processmining.xesalignmentextension.XAlignmentExtension;
import org.processmining.xesalignmentextension.XAlignmentExtension.XAlignedLog;

import com.google.common.base.Function;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Maps;

/**
 * Supervised Log Abstraction based on activity patterns. *
 * 
 * @author F. Mannhardt
 *
 */
@Plugin(name = "Log Abstraction - Abstract Log based on Patterns", parameterLabels = { "Log",
		"Abstraction Patterns" }, returnLabels = { "Abstracted Log" }, returnTypes = {
				XLog.class }, userAccessible = true, categories = { PluginCategory.Enhancement }, keywords = {
						"abstraction", "supervised",
						"pattern" }, help = "<html>Abstracts events from the supplied event log using a supervised pattern-based approach. "
								+ "This plug-in implements the methodology presented in the BPM'16 paper 'From Low-Level Events to Activities - A Pattern-based Approach' (DOI: <a href=\"http://dx.doi.org/10.1007/978-3-319-45348-4_8\">10.1007/978-3-319-45348-4_8)</a>."
								+ "Activity patterns describe the behavior expected for an (abstract) activity and should be modeled as Data Petri Net (DPN) with a single source and a single sink place (SESE stucture).</html>")
public final class PatternBasedLogAbstractionPlugin {

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "LogEnhancement")
	@PluginVariant(variantLabel = "Log Abstraction - Abstract Log based on Patterns", requiredParameterLabels = { 0,
			1 })
	public XLog abstractLogBasedOnPatterns(UIPluginContext context, XLog log, Petrinet... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {
		return abstractLogBasedOnPatterns(context, log, (PetrinetGraph[]) abstractionNets);
	}

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl", pack = "LogEnhancement")
	@PluginVariant(variantLabel = "Log Abstraction - Abstract Log based on Patterns", requiredParameterLabels = { 0,
			1 })
	public XLog abstractLogBasedOnPatterns(UIPluginContext context, XLog log, DataPetriNet... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {
		return abstractLogBasedOnPatterns(context, log, (PetrinetGraph[]) abstractionNets);
	}

	public XLog abstractLogBasedOnPatterns(UIPluginContext context, XLog log, PetrinetGraph... abstractionNets)
			throws ControlFlowAlignmentException, DataAlignmentException, ParseException, CompositionVisitorException,
			PatternStructureException {

		try {
			InteractionResult result = context
					.showConfiguration("Abstract Low-Level Events to Activities using Patterns",
							new JLabel("<html>"
									+ "This plug-in implements the methodology presented in the BPM'16 paper 'From Low-Level Events to Activities - A Pattern-based Approach': (DOI: <a href=\"http://dx.doi.org/10.1007/978-3-319-45348-4_8\">10.1007/978-3-319-45348-4_8)</a>.<br/> "
									+ "The required input is:<br/>" + "<ul>" + "<li>an event log;</li>"
									+ "<li>activity patterns modeled as Data Petri Net (DPN).</li>" + "</ul>"
									+ "Through an alignment between those activity patterns and the event log an abstracted event log is obtained.<br/>"
									+ "Events in this abstracted event log correspond to instantiations of recognizable activities.<br/> There are some assumptions made on the input:"
									+ "<ul>" + "<li>Each input DPN has a single source and a single sink place.</li>"
									+ "<li>Used variables have unique names across all DPNs.</li>" + "</ul>"
									+ "<p>In the following dialogs you are asked for some configuration options.<br/> When in doubt, use the pre-set <b>default options</b>.</p>"
									+ "</html>"));
			if (result == InteractionResult.CANCEL) {
				throw new UserCancelledException();
			}

			DataPetriNet[] transformedNets = transformNets(abstractionNets);

			Map<DataPetriNet, String> patternToName = ProMUIHelper.queryMapToString(context,
					"Configure the mapping between activity patterns (left) and the desired names for the abstracted events (right)",
					transformedNets,
					Maps.asMap(ImmutableSet.copyOf(transformedNets), new Function<DataPetriNet, String>() {

						public String apply(DataPetriNet net) {
							return net.getLabel().replaceAll(".pnml", "");
						}
					}));

			Map<String, AbstractionPattern> abstractionPatterns = buildPatterns(patternToName);

			boolean keepUnmappedEvents = JOptionPane.showConfirmDialog(null,
					"<html>Do you want to keep unmapped low-level events that are not part of any activity pattern?<br/>"
							+ "This option can be useful if your event log already contains some events at the desired level of abstraction."
							+ "</html>",
					"Configure how to deal with unmapped events", JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION;

			String initialModel = buildInitialModel(abstractionPatterns);
			AbstractionModelBuilder modelBuilder = new AbstractionModelBuilder(abstractionPatterns);
			modelBuilder.update(initialModel);

			result = context.showConfiguration("Configure the abstraction model (restrict the possible interaction)",
					modelBuilder);
			while (!modelBuilder.isValid() && result != InteractionResult.CANCEL) {
				result = context.showConfiguration("Invalid Model: Specify a valid abstraction model!", modelBuilder);
			}
			if (result != InteractionResult.CANCEL) {
				return abstractPatterns(context, log, modelBuilder.getAbstractionModel(), keepUnmappedEvents);
			} else {
				context.getFutureResult(0).cancel(false);
				return null;
			}
		} catch (UserCancelledException e) {
			context.getFutureResult(0).cancel(false);
			return null;
		}
	}

	public static XLog abstractPatterns(final UIPluginContext context, XLog log, AbstractionModel model,
			boolean keepUnmappedEvents) throws UserCancelledException, ControlFlowAlignmentException,
			DataAlignmentException, PatternStructureException {

		BalancedProcessorConfiguration config = queryConfiguration(context, model, log);

		// Usually the search space will be huge, so rather waste CPU cycles here instead of memory
		config.setActivateDataViewCache(false);
		config.setKeepDataFlowSearchSpace(false);

		config.setObserver(new DataConformancePlusObserverImpl(context) {

			public void startAlignment(int numExpectedResults) {
				super.startAlignment(numExpectedResults);
				context.getProgress().setMaximum(context.getProgress().getMaximum() + 1);
			}

		});

		XAlignedLog alignedLog = alignLogToAbstractionModel(context.getProgress(), config, log, model);
		LogAbstracterBasedOnAlignment abstractionProcessor = createAbstractionProcessor(model,
				config.getActivityMapping());

		if (reviewAlignment(context, alignedLog, model.getPatterns()) != InteractionResult.CANCEL) {

			ProvidedObjectHelper.publish(context, "Abstraction: Alignment", alignedLog.getLog(), XLog.class, false);
			ProvidedObjectHelper.publish(context, "Abstraction: Model", model.getCombinedDPN(), DataPetriNet.class,
					false);

			double errorRateLimit = 1.0d;
			try {
				errorRateLimit = ProMUIHelper.queryForDouble(context,
						"Specify 'matching error' ([0.0,1.0], empty => use all matches)");
			} catch (NumberFormatException e) {
				// means 1.0d
			}

			try {
				abstractionProcessor.setErrorRateLimit(errorRateLimit);
				abstractionProcessor.setKeepUnmappedEvents(keepUnmappedEvents);
				return abstractionProcessor.abstractBasedOnPatterns(alignedLog);
			} catch (PatternStructureException e) {
				throw e;
			}

		} else {
			throw new UserCancelledException();
		}
	}

	public static Map<String, AbstractionPattern> buildPatterns(Map<DataPetriNet, String> patternToName) {
		Map<String, AbstractionPattern> abstractionPatterns = new HashMap<>();
		for (Entry<DataPetriNet, String> entry : patternToName.entrySet()) {
			AbstractionPattern pattern = new AbstractionPattern(entry.getValue(), entry.getKey());
			pattern.setStartTransition(findStartTransition(pattern.getDPN()));
			pattern.setCompleteTransition(findCompleteTransition(pattern.getDPN()));
			for (Transition t : entry.getKey().getTransitions()) {
				pattern.getTransitions().add(t);
			}
			abstractionPatterns.put(pattern.getPatternName(), pattern);
		}
		return abstractionPatterns;
	}

	public static AbstractionModel composePatterns(String abstractionModel,
			Map<String, AbstractionPattern> abstractionPatterns, boolean simplify)
			throws ParseException, CompositionVisitorException {
		CompositionRoot compositionModel;
		if (simplify) {
			compositionModel = PatternSimplifier.simplify(abstractionModel);
		} else {
			compositionModel = new CompositionParser(abstractionModel).parse();
		}
		return PatternCompiler.compile(compositionModel, "Composed Model", abstractionPatterns, simplify);
	}

	@Deprecated
	public static XLog abstractPatterns(PluginContext context, XLog log, AbstractionModel abstractionModel,
			boolean keepUnmappedEvents, double errorRateLimit, BalancedProcessorConfiguration alignmentConfig)
			throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException {
		return abstractPatterns(context.getProgress(), log, abstractionModel, keepUnmappedEvents, errorRateLimit,
				alignmentConfig);
	}

	public static XLog abstractPatterns(Progress progress, XLog log, AbstractionModel abstractionModel,
			boolean keepUnmappedEvents, double errorRateLimit, BalancedProcessorConfiguration alignmentConfig)
			throws ControlFlowAlignmentException, DataAlignmentException, PatternStructureException {

		XAlignedLog alignedLog = alignLogToAbstractionModel(progress, alignmentConfig, log, abstractionModel);
		return abstractAlignedLog(abstractionModel, alignedLog, keepUnmappedEvents, errorRateLimit,
				alignmentConfig.getActivityMapping());
	}

	public static XLog abstractAlignedLog(AbstractionModel abstractionModel, XAlignedLog alignedLog,
			boolean keepUnmappedEvents, double errorRateLimit, TransEvClassMapping transitionMapping)
			throws PatternStructureException {
		LogAbstracterBasedOnAlignment abstractionProcessor = createAbstractionProcessor(abstractionModel,
				transitionMapping);
		abstractionProcessor.setErrorRateLimit(errorRateLimit);
		abstractionProcessor.setKeepUnmappedEvents(keepUnmappedEvents);
		return abstractionProcessor.abstractBasedOnPatterns(alignedLog);
	}

	public static XAlignedLog alignLogToAbstractionModel(final Progress progress,
			BalancedProcessorConfiguration alignmentConfig, XLog log, AbstractionModel model)
			throws ControlFlowAlignmentException, DataAlignmentException {
		BalancedDataXAlignmentPlugin alignmentPlugin = new BalancedDataXAlignmentPlugin();
		XLog alignedLog = alignmentPlugin.alignLog(progress, model.getCombinedDPN(), log, alignmentConfig);
		return XAlignmentExtension.instance().extendLog(alignedLog);
	}

	public static LogAbstracterBasedOnAlignment createAbstractionProcessor(AbstractionModel abstractionModel,
			TransEvClassMapping transitionMapping) throws PatternStructureException {

		BiMap<Transition, String> startPoints = HashBiMap.create();
		BiMap<Transition, String> endPoints = HashBiMap.create();
		//TODO support more than 'start' and 'complete'
		for (AbstractionPattern pattern : abstractionModel.getPatterns()) {
			// Translate all transitions
			pattern.setStartTransition((Transition) abstractionModel.getOld2NewMap().get(pattern.getStartTransition()));
			pattern.setCompleteTransition(
					(Transition) abstractionModel.getOld2NewMap().get(pattern.getCompleteTransition()));
			ImmutableList<Transition> oldTransitions = ImmutableList.copyOf(pattern.getTransitions());
			pattern.getTransitions().clear();
			for (Transition t : oldTransitions) {
				pattern.getTransitions().add((Transition) abstractionModel.getOld2NewMap().get(t));
			}
			// Configure abstraction with life-cycle transitions
			startPoints.put(pattern.getStartTransition(), pattern.getPatternName());
			endPoints.put(pattern.getCompleteTransition(), pattern.getPatternName());
		}

		return new LogAbstracterBasedOnAlignment(startPoints, endPoints, transitionMapping);
	}

	public static DataPetriNet[] transformNets(PetrinetGraph[] abstractionNets) {
		DataPetriNet[] transformedNets = new DataPetriNet[abstractionNets.length];
		for (int i = 0; i < abstractionNets.length; i++) {
			transformedNets[i] = transformNet(abstractionNets[i]);
		}
		return transformedNets;
	}

	public static DataPetriNet transformNet(PetrinetGraph graph) {
		PetriNetWithDataFactory factory = new PetriNetWithDataFactory(graph, graph.getLabel());
		PetriNetWithData transformedNet = factory.getRetValue();

		if (graph instanceof DataPetriNetsWithMarkings) {
			DataPetriNetsWithMarkings dpn = (DataPetriNetsWithMarkings) graph;
			if (dpn.getInitialMarking() != null) {
				transformedNet.setInitialMarking(factory.convertMarking(dpn.getInitialMarking()));
			}
			if (dpn.getFinalMarkings() != null) {
				Marking[] finalMarkings = dpn.getFinalMarkings();
				if (finalMarkings != null && finalMarkings.length > 0 && finalMarkings[0] != null) {
					transformedNet.setFinalMarkings(new Marking[] { factory.convertMarking(finalMarkings[0]) });
				}
			}
		}

		Place source = findSource(transformedNet);
		if (source == null) {
			throw new IllegalArgumentException("Invalid pattern: Cannot find unique source place!");
		}
		// choice after source place or loop back to source place requires invisible transition
		if (transformedNet.getOutEdges(source).size() > 1 || transformedNet.getInEdges(source).size() > 0) {
			Transition artificalStart = transformedNet.addTransition("artificalStart");
			artificalStart.setInvisible(true);
			Place artificalSource = transformedNet.addPlace("artificalSource");
			transformedNet.addArc(artificalSource, artificalStart);
			transformedNet.addArc(artificalStart, source);
		}

		Place sink = findSink(transformedNet);
		if (sink == null) {
			throw new IllegalArgumentException("Invalid pattern: Cannot find unique sink place!");
		}
		// choice before sink place or loop back from sink place requires invisible transition
		if (transformedNet.getInEdges(sink).size() > 1 || transformedNet.getOutEdges(source).size() > 0) {
			Transition artificalEnd = transformedNet.addTransition("artificalEnd");
			artificalEnd.setInvisible(true);
			Place artificalSink = transformedNet.addPlace("artificalSink");
			transformedNet.addArc(sink, artificalEnd);
			transformedNet.addArc(artificalEnd, artificalSink);
		}

		if (graph instanceof DataPetriNet) {
			Map<DataElement, DataElement> oldVar2New = new HashMap<>();
			for (DataElement element : ((DataPetriNet) graph).getVariables()) {
				DataElement newVar = transformedNet.addVariable(element.getVarName(), element.getType(),
						element.getMinValue(), element.getMaxValue());
				oldVar2New.put(element, newVar);
			}
			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : graph.getEdges()) {
				if (edge instanceof VariableAccess) {
					if (edge.getSource() instanceof Transition) {
						DataElement oldVar = (DataElement) edge.getTarget();
						DataElement newVar = oldVar2New.get(oldVar);
						transformedNet.assignWriteOperation(factory.getTransMapping().get(edge.getSource()), newVar);
					}
					if (edge.getTarget() instanceof Transition) {
						DataElement oldVar = (DataElement) edge.getSource();
						DataElement newVar = oldVar2New.get(oldVar);
						transformedNet.assignReadOperation(factory.getTransMapping().get(edge.getSource()), newVar);
					}
				}
			}
			for (Transition t : graph.getTransitions()) {
				PNWDTransition transition = (PNWDTransition) factory.getTransMapping().get(t);
				if (((PNWDTransition) t).hasGuardExpression()) {
					try {
						transition.setGuard(transformedNet, ((PNWDTransition) t).getGuardExpression());
					} catch (NonExistingVariableException e) {
						throw new RuntimeException(e);
					}
				}
			}
		}

		return transformedNet;
	}

	private static Place findSink(PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking[] finalMarkings = ((DataPetriNetsWithMarkings) net).getFinalMarkings();
			if (finalMarkings != null && finalMarkings.length > 0) {
				Marking finalMarking = ((DataPetriNetsWithMarkings) net).getFinalMarkings()[0];
				if (finalMarking != null) {
					if (finalMarking.size() == 1) {
						return finalMarking.iterator().next();
					} else {
						return null; // more than one place marked TODO
					}
				}
			}
		}
		// Marking did not work try structure
		return findSinkByStructure(net);
	}

	private static Place findSinkByStructure(PetrinetGraph net) {
		Place sink = null;
		for (Place p : net.getPlaces()) {
			if (net.getOutEdges(p).isEmpty()) {
				if (sink == null) {
					sink = p;
				} else {
					return null;
				}
			}
		}
		return sink;
	}

	private static Place findSource(PetrinetGraph net) {
		if (net instanceof DataPetriNetsWithMarkings) {
			Marking initialMarking = ((DataPetriNetsWithMarkings) net).getInitialMarking();
			if (initialMarking != null) {
				if (initialMarking.size() == 1) {
					return initialMarking.iterator().next();
				} else {
					return null; // more than one place marked TODO
				}
			}
		}
		// Marking did not work try structure
		return findSourceByStructure(net);
	}

	private static Place findSourceByStructure(PetrinetGraph net) {
		Place source = null;
		for (Place p : net.getPlaces()) {
			if (net.getInEdges(p).isEmpty()) {
				if (source == null) {
					source = p;
				} else {
					return null;
				}
			}
		}
		return source;
	}

	public static String buildInitialModel(Map<String, AbstractionPattern> abstractionPatterns) {
		StringBuilder sb = new StringBuilder();
		for (Iterator<String> iter = abstractionPatterns.keySet().iterator(); iter.hasNext();) {
			String id = iter.next();
			if (!SourceVersion.isIdentifier(id)) {
				id = "\"" + id + "\"";
			}
			sb.append(id);
			sb.append("*");
			if (iter.hasNext()) {
				sb.append('#');
			}
		}
		return sb.toString();
	}

	private static BalancedProcessorConfiguration queryConfiguration(UIPluginContext context,
			AbstractionModel abstractionModel, XLog log) throws UserCancelledException {

		DataPetriNetsWithMarkings net = abstractionModel.getCombinedDPN();

		BalancedProcessorConfiguration config = DataConformanceConfigUIUtils.queryBalancedAlignmentConfig(context);

		TransEvClassMapping activityMapping = DataConformanceConfigUIUtils.queryActivityMapping(context, net, log);

		config.setActivityMapping(activityMapping);
		config.setInitialMarking(net.getInitialMarking());
		config.setFinalMarkings(net.getFinalMarkings());

		DataConformanceConfigUIUtils.queryControlFlowCost(context, net, log, config, 1, 1);

		if (!net.getVariables().isEmpty()) {
			DataConformanceConfigUIUtils.queryDataAlignmentConfig(context, net, log, activityMapping, config);
		} else {

			VariableMatchCosts variableCost = BalancedProcessorConfiguration.createDefaultVariableCost(net,
					Collections.<String>emptySet(), 0, 0);
			config.setVariableCost(variableCost);
			config.setVariableMapping(Collections.<String, String>emptyMap());

		}
		return config;
	}

	private static InteractionResult reviewAlignment(final UIPluginContext context, XAlignedLog alignedLog,
			Collection<AbstractionPattern> patterns) {
		AlignmentGuidedPatternResult result = new AlignmentGuidedPatternResult(alignedLog, patterns);
		return context.showConfiguration(String.format("Review alignment (%s matches, %.2f error rate)",
				result.getNumMatches(), result.getErrorRate()), result.getPanel());
	}

	private static Transition findCompleteTransition(PetrinetGraph net) {
		Place sink = findSinkByStructure(net);
		if (net.getInEdges(sink).size() != 1) {
			throw new IllegalArgumentException("Invalid pattern: There is more than 1 complete transition!");
		} else {
			return (Transition) net.getInEdges(sink).iterator().next().getSource();
		}
	}

	private static Transition findStartTransition(PetrinetGraph net) {
		Place source = findSourceByStructure(net);
		if (net.getOutEdges(source).size() != 1) {
			throw new IllegalArgumentException("Invalid pattern: There is more than 1 start transition!");
		} else {
			return (Transition) net.getOutEdges(source).iterator().next().getTarget();
		}
	}
}