package org.processmining.logenhancement.abstraction.editor;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Map;
import java.util.concurrent.ExecutionException;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JEditorPane;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingWorker;
import javax.swing.Timer;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultHighlighter;
import javax.swing.text.Highlighter;

import org.fife.ui.rtextarea.RTextScrollPane;
import org.processmining.datapetrinets.dsl.editor.AutoCompletingRTextArea;
import org.processmining.datapetrinets.expression.syntax.ParseException;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverter.DPNAsDot;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.GuardDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.PlaceDisplayMode;
import org.processmining.datapetrinets.visualization.graphviz.DPNGraphvizConverterPlugin.VariableDisplayMode;
import org.processmining.logenhancement.abstraction.model.AbstractionModel;
import org.processmining.logenhancement.abstraction.model.AbstractionPattern;
import org.processmining.logenhancement.abstraction.model.PatternCompiler;
import org.processmining.logenhancement.abstraction.model.PatternSimplifier;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParser;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRoot;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.plugins.graphviz.dot.Dot;
import org.processmining.plugins.graphviz.dot.Dot.GraphDirection;
import org.processmining.plugins.graphviz.visualisation.DotPanel;

import com.fluxicon.slickerbox.factory.SlickerFactory;
import com.google.common.base.Throwables;
import com.kitfox.svg.SVGDiagram;

public class AbstractionModelBuilder extends JPanel {
	
	private static final String SYNTAX_HELP_TEXT = "<HTML><h1>Model Syntax</h1>"
			+ "Initially all abstraction patterns are composed in parallel. "
			+ "The composition can be changed by manipulating the formula or by using the context menu (right-click) of the model."
			+ "Operators can be nested, patterns can appear multiple times and parentheses can be used to indicate preceedence if necessary."
			+ "<h2>Examples for patterns A and B</h2>"
			+ "<ul>"
			+ "<li>'A#B' - A and B can happen in parallel</li>"
			+ "<li>'A|B' - Either A or B can happen</li>"
			+ "<li>'A<B' - A must happen before B</li>"
			+ "<li>'%[A,B]' - A and B can interleave</li>"
			+ "<li>'A?' - A can happen 1 or 0 times</li>"
			+ "<li>'A+' - A can happen 1 or multiple times</li>"
			+ "<li>'A*' - A can happen 0 or multiple times</li>"
			+ "<li>'A{n,m}' - A can happen n to m times</li>"
			+ "</ul>"
			+ "<h2>Full Syntax</h2>"
			+ "<pre>"
			+ "parse		:=	expression <EOF>\r\n" + 
			"expression	:=	parallel ( ( \"#\" parallel ) )*\r\n" + 
			"parallel	:=	choice ( ( \"|\" choice ) )*\r\n" + 
			"choice		:=	sequence ( ( \"<\" sequence ) )*\r\n" + 
			"sequence	:=	( \"%[\" nary ( \",\" nary )+ \"]\" )\r\n" + 
			"		|	( \"|[\" nary ( \",\" nary )+ \"]\" )\r\n" + 
			"		|	( \"#[\" nary ( \",\" nary )+ \"]\" )\r\n" + 
			"		|	nary\r\n" + 
			"nary		:=	basic ( ( \"*\" ) | ( \"+\" ) | ( \"?\" ) | ( \"{\" <INTEGER> \",\" ( <INTEGER> | \"*\" ) \"}\" ) )?\r\n" + 
			"basic		:=	( \"(\" expression \")\" )\r\n" + 
			"		|	identifier\r\n" + 
			"identifier	:=	( <IDENTIFIER> | <STRING_LITERAL> )"		
			+ "</pre></HTML>";

	private interface DotAndSVG {

		Dot getDot();

		SVGDiagram getSVG();

	}

	public enum Mode {
		ABSTRACTION_MODEL, DPN_BASIC, DPN_FULL
	}

	private class CompositionModelParserWorker extends SwingWorker<DotAndSVG, Void> {

		private final String currentCode;

		private CompositionModelParserWorker(String currentText) {
			this.currentCode = currentText;
		}

		@Override
		protected DotAndSVG doInBackground() throws Exception {
			AbstractionModel model = compileAbstractionModel(currentCode);
			if (mode == Mode.ABSTRACTION_MODEL) {
				final AbstractionModelAsDot modelWithDot = DotCompiler.convert(model);
				return new DotAndSVG() {

					public Dot getDot() {
						return modelWithDot.getDot();
					}

					public SVGDiagram getSVG() {
						return modelWithDot.getSVG();
					}
				};
			} else if (mode == Mode.DPN_BASIC) {
				final DPNAsDot dpnAsDot = DPNGraphvizConverter.convertDPN(model.getCombinedDPN(), model
						.getCombinedDPN().getInitialMarking(), model.getCombinedDPN().getFinalMarkings(),
						GuardDisplayMode.HIDDEN, VariableDisplayMode.HIDDEN, PlaceDisplayMode.BASIC,
						GraphDirection.leftRight);
				final SVGDiagram svg = DotPanel.dot2svg(dpnAsDot.getDot());
				return new DotAndSVG() {

					public SVGDiagram getSVG() {
						return svg;
					}

					public Dot getDot() {
						return dpnAsDot.getDot();
					}
				};
			} else if (mode == Mode.DPN_FULL) {
				final DPNAsDot dpnAsDot = DPNGraphvizConverter.convertDPN(model.getCombinedDPN(), model
						.getCombinedDPN().getInitialMarking(), model.getCombinedDPN().getFinalMarkings(),
						GuardDisplayMode.EDGES, VariableDisplayMode.NO_LAYOUT, PlaceDisplayMode.BASIC,
						GraphDirection.leftRight);
				final SVGDiagram svg = DotPanel.dot2svg(dpnAsDot.getDot());
				return new DotAndSVG() {

					public SVGDiagram getSVG() {
						return svg;
					}

					public Dot getDot() {
						return dpnAsDot.getDot();
					}
				};
			} else {
				throw new IllegalArgumentException("Unkown mode " + mode);
			}
		}

		@Override
		protected void done() {
			try {
				if (currentErrorHighlight != null) {
					textArea.getHighlighter().removeHighlight(currentErrorHighlight);
					currentErrorHighlight = null;
				}

				final DotAndSVG result = get();
				previewPanel.changeDot(result.getDot(), result.getSVG(), true);

				infoMessages.setText("Updated Abstraction Model successfully!");
				isValid = true;

			} catch (ExecutionException e) {
				isValid = false;
				if (e.getCause() instanceof ParseException) {
					ParseException parseEx = (ParseException) e.getCause();
					int beginLine = parseEx.currentToken.next.beginLine - 1;
					int beginColumn = parseEx.currentToken.next.beginColumn - 1;
					int endLine = parseEx.currentToken.next.endLine - 1;
					int endColumn = parseEx.currentToken.next.endColumn - 1;
					try {
						int beginOffset = textArea.getLineStartOffset(beginLine);
						int endOffset = textArea.getLineStartOffset(endLine);
						currentErrorHighlight = textArea.getHighlighter().addHighlight(beginOffset + beginColumn,
								endOffset + endColumn, highlightPainter);
					} catch (BadLocationException e1) {
					}
					infoMessages.setText("Error updating DPN: " + parseEx.getMessage());
					infoMessages.setCaretPosition(0);
				} else {
					infoMessages.setText("Error updating DPN: " + Throwables.getRootCause(e).getMessage());
					infoMessages.setCaretPosition(0);
				}
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
			} finally {
				updateFinished();
			}
		}

	}

	private static final long serialVersionUID = 1L;
	private static final int ACTION_DELAY = 1000;
	private static final String COMPOSED_MODEL = "ComposedModel";

	private final Map<String, AbstractionPattern> abstractionPatterns;

	private final AutoCompletingRTextArea textArea;
	private final DotPanel previewPanel;
	private final JTextArea infoMessages;

	private boolean isValid;
	private final Timer updateTimer;

	private Highlighter.HighlightPainter highlightPainter;
	private Object currentErrorHighlight;
	protected Mode mode = Mode.ABSTRACTION_MODEL;
	protected boolean isSimplify = true;

	public AbstractionModelBuilder(Map<String, AbstractionPattern> abstractionPatterns) {
		super();
		this.abstractionPatterns = abstractionPatterns;

		setLayout(new BorderLayout());
		setBackground(Color.WHITE);
		setOpaque(true);
		setPreferredSize(new Dimension(600, 500));
		
		JPanel topPanel = new JPanel();
		topPanel.setLayout(new BoxLayout(topPanel, BoxLayout.X_AXIS));
		topPanel.setBackground(Color.WHITE);
		topPanel.setOpaque(true);
		
		textArea = new AutoCompletingRTextArea();
		textArea.setHighlightCurrentLine(false);
		textArea.setLineWrap(true);

		highlightPainter = new DefaultHighlighter.DefaultHighlightPainter(Color.RED);
		updateTimer = new Timer(ACTION_DELAY, new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				doUpdate(textArea.getText());
				updateTimer.stop();
			}

		});
		textArea.setToolTipText(SYNTAX_HELP_TEXT);

		RTextScrollPane scrollPane = new RTextScrollPane(textArea);
		scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
		scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED);
		scrollPane.setPreferredSize(new Dimension(Short.MAX_VALUE, 80));
		topPanel.add(scrollPane);
		
		JButton helpButton = SlickerFactory.instance().createButton("?");
		helpButton.setMinimumSize(new Dimension(30, 30));
		helpButton.setPreferredSize(new Dimension(30, 30));
		helpButton.setMaximumSize(new Dimension(30, 30));
		helpButton.addActionListener(new ActionListener() {
			
			public void actionPerformed(ActionEvent e) {
				JEditorPane pane = new JEditorPane("text/html", SYNTAX_HELP_TEXT);
				pane.setEditable(false);
				JOptionPane.showConfirmDialog(textArea, pane, "Syntax Help", JOptionPane.OK_CANCEL_OPTION, JOptionPane.INFORMATION_MESSAGE);
			}
			
		});
		topPanel.add(helpButton);
		add(topPanel, BorderLayout.NORTH);
		
		textArea.getDocument().addDocumentListener(new DocumentListener() {

			public void removeUpdate(DocumentEvent e) {
				scheduleUpdate();
			}

			public void insertUpdate(DocumentEvent e) {
				if (e.getLength() == e.getDocument().getLength()) {
					try {
						doUpdate(e.getDocument().getText(0, e.getDocument().getLength()));
					} catch (BadLocationException e1) {
						scheduleUpdate();
					}
				} else {
					scheduleUpdate();
				}
			}

			public void changedUpdate(DocumentEvent e) {
				scheduleUpdate();
			}
		});
		Dot dot = new Dot();
		dot.addNode("loading ...");
		previewPanel = new DotPanel(dot);
		previewPanel.setDirection(GraphDirection.leftRight);
		previewPanel.setBackground(Color.WHITE);
		previewPanel.setOpaque(true);
		previewPanel.setPreferredSize(new Dimension(Short.MAX_VALUE, Short.MAX_VALUE));
		add(previewPanel, BorderLayout.CENTER);

		JPanel bottomPanel = new JPanel();
		bottomPanel.setLayout(new BoxLayout(bottomPanel, BoxLayout.X_AXIS));
		bottomPanel.setBackground(Color.WHITE);
		bottomPanel.setOpaque(true);

		infoMessages = new JTextArea();
		JScrollPane scrollPane2 = new JScrollPane(infoMessages);
		scrollPane2.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
		scrollPane2.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		bottomPanel.add(scrollPane2);

		JPanel configPanel = new JPanel();
		configPanel.setLayout(new BoxLayout(configPanel, BoxLayout.Y_AXIS));
		configPanel.setBackground(Color.WHITE);
		configPanel.setOpaque(true);
		
		final JComboBox<Mode> showDPN = new JComboBox<Mode>(Mode.values());
		showDPN.setBackground(Color.WHITE);
		showDPN.setOpaque(true);
		showDPN.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				mode = (Mode) showDPN.getSelectedItem();
				update(textArea.getText());
			}
		});
		showDPN.setMinimumSize(new Dimension(200, 20));
		showDPN.setMaximumSize(new Dimension(200, 20));
		showDPN.setAlignmentX(CENTER_ALIGNMENT);
		configPanel.add(showDPN);
		
		final JCheckBox simplifyModel = new JCheckBox("Simplify Model", true);
		simplifyModel.setBackground(Color.WHITE);
		simplifyModel.setOpaque(true);
		simplifyModel.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				isSimplify = simplifyModel.getModel().isSelected();
				update(textArea.getText());
			}
		});
		simplifyModel.setMinimumSize(new Dimension(100, 20));
		simplifyModel.setMaximumSize(new Dimension(100, 20));
		simplifyModel.setAlignmentX(CENTER_ALIGNMENT);
		configPanel.add(simplifyModel);
		
		bottomPanel.add(configPanel);
		add(bottomPanel, BorderLayout.SOUTH);
	}

	public void update(String newText) {
		textArea.setText(newText);
		doUpdate(newText);
	}

	private void doUpdate(String newText) {
		updateStarted();
		new CompositionModelParserWorker(newText).execute();
	}

	public void updateStarted() {
		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	}

	public void updateFinished() {
		setCursor(Cursor.getDefaultCursor());
	}

	protected void scheduleUpdate() {
		updateTimer.restart();
	}

	public AbstractionModel getAbstractionModel()
			throws org.processmining.logenhancement.abstraction.model.syntax.ParseException,
			CompositionVisitorException {
		return compileAbstractionModel(textArea.getText());
	}

	public AbstractionModel compileAbstractionModel(String code)
			throws org.processmining.logenhancement.abstraction.model.syntax.ParseException,
			CompositionVisitorException {
		CompositionRoot compositionModel;
		if (isSimplify) {
			compositionModel = PatternSimplifier.simplify(code);
		} else {
			compositionModel = new CompositionParser(code).parse();
		}
		return PatternCompiler.compile(compositionModel, COMPOSED_MODEL, abstractionPatterns, isSimplify);
	}

	public boolean isValid() {
		return isValid;
	}

}
