package org.processmining.logenhancement.abstraction.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.datapetrinets.DataPetriNetsWithMarkings;
import org.processmining.datapetrinets.exception.NonExistingVariableException;
import org.processmining.datapetrinets.expression.GuardExpression;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionCho;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionId;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionInt;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionNCho;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionNPar;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionPar;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParser;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParserVisitor;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepExact;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepInf;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepInfOne;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepZeroOne;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRoot;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionSeq;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.logenhancement.abstraction.model.syntax.SimpleNode;
import org.processmining.models.graphbased.directed.AbstractDirectedGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetEdge;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.PetrinetNode;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.DataElement;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PNWDTransition;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.semantics.petrinet.Marking;

import com.google.common.base.Predicate;
import com.google.common.collect.Collections2;

public class PatternCompiler {

	private interface BoundaryNodes {
		Place getSource();

		Place getSink();
	}

	private static final class BoundaryNodesImpl implements BoundaryNodes {
		private final Place source;
		private final Place sink;

		private BoundaryNodesImpl(Place source, Place sink) {
			this.source = source;
			this.sink = sink;
		}

		public Place getSource() {
			return source;
		}

		public Place getSink() {
			return sink;
		}
	}

	private static final class LabelGenerator {

		int tCount = 0;
		int pCount = 0;
		int vCount = 0;

		public String getPlaceName(PetrinetGraph parentModel, Place place) {
			return place.getLabel();
		}

		public String getTransitionName(PetrinetGraph parentModel, Transition transition) {
			return transition.getLabel();
		}

		public String getAnonymousPlaceName() {
			return "p".concat(String.valueOf(pCount++));
		}

		public String getAnonymousTransitionName() {
			return "t".concat(String.valueOf(tCount++));
		}

		public String getAnonymousVariableName() {
			return "v".concat(String.valueOf(vCount++));
		}

	}

	private static final class CompilingVisitor implements CompositionParserVisitor {

		private final DataPetriNetsWithMarkings net;
		private final Map<String, AbstractionPattern> patternMap;
		private final LabelGenerator labelGenerator;
		private final Map<PetrinetNode, PetrinetNode> old2New;
		private final boolean reduceTau;

		public CompilingVisitor(String netName, Map<String, AbstractionPattern> patternMap, boolean reduceTau) {
			this.reduceTau = reduceTau;
			this.net = new PetriNetWithData(netName);
			this.patternMap = patternMap;
			this.labelGenerator = new LabelGenerator();
			this.old2New = new HashMap<>();
		}

		public Object visit(SimpleNode node, Object data) throws CompositionVisitorException {
			throw new IllegalStateException("No unamed nodes allowed!");
		}

		public Object visit(CompositionRoot node, Object data) throws CompositionVisitorException {
			if (node.jjtGetNumChildren() > 1) {
				throw new IllegalArgumentException("Invalid expression " + node
						+ " should not have been parsed! Top level element is only allowed to have one child.");
			}
			BoundaryNodes boundaryNodes = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			setMarking(boundaryNodes);
			return net;
		}

		private void setMarking(BoundaryNodes boundaryNodes) {
			Marking initialMarking = new Marking();
			initialMarking.add(boundaryNodes.getSource());
			net.setInitialMarking(initialMarking);
			Marking finalMarking = new Marking();
			finalMarking.add(boundaryNodes.getSink());
			net.setFinalMarkings(new Marking[] { finalMarking });
		}

		public Object visit(CompositionPar node, Object data) throws CompositionVisitorException {
			BoundaryNodes boundaryNodes1 = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			BoundaryNodes boundaryNodes2 = (BoundaryNodes) node.jjtGetChild(1).jjtAccept(this, data);
			return composeInParallel(boundaryNodes1, boundaryNodes2);
		}

		private BoundaryNodes composeInParallel(BoundaryNodes boundaryNodes1, BoundaryNodes boundaryNodes2) {
			Place source = addAnonymousPlace();
			Transition split = addInvisibleTransition();
			Transition merge = addInvisibleTransition();
			Place sink = addAnonymousPlace();
			net.addArc(source, split);
			net.addArc(split, boundaryNodes1.getSource());
			net.addArc(split, boundaryNodes2.getSource());
			net.addArc(boundaryNodes1.getSink(), merge);
			net.addArc(boundaryNodes2.getSink(), merge);
			net.addArc(merge, sink);
			return new BoundaryNodesImpl(source, sink);
		}

		public Object visit(CompositionNPar node, Object data) throws CompositionVisitorException {
			Collection<BoundaryNodes> patterns = new ArrayList<>();
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				patterns.add((BoundaryNodes) node.jjtGetChild(i).jjtAccept(this, data));
			}
			return composeInNaryParallel(patterns);
		}

		private Object composeInNaryParallel(Collection<BoundaryNodes> patternBoundaries) {
			Place source = addAnonymousPlace();
			Transition split = addInvisibleTransition();
			Transition merge = addInvisibleTransition();
			Place sink = addAnonymousPlace();
			for (BoundaryNodes boundary : patternBoundaries) {
				net.addArc(split, boundary.getSource());
				net.addArc(boundary.getSink(), merge);
			}
			net.addArc(source, split);
			net.addArc(merge, sink);
			return new BoundaryNodesImpl(source, sink);
		}

		public Object visit(CompositionCho node, Object data) throws CompositionVisitorException {
			BoundaryNodes boundaryNodes1 = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			BoundaryNodes boundaryNodes2 = (BoundaryNodes) node.jjtGetChild(1).jjtAccept(this, data);
			return composeInChoice(boundaryNodes1, boundaryNodes2);
		}

		private BoundaryNodes composeInChoice(BoundaryNodes boundaryNodes1, BoundaryNodes boundaryNodes2) {
			Place source = addAnonymousPlace();
			connectEliminateTauForward(source, boundaryNodes1.getSource());
			connectEliminateTauForward(source, boundaryNodes2.getSource());
			Place sink = addAnonymousPlace();
			connectEliminateTauBackward(boundaryNodes1.getSink(), sink);
			connectEliminateTauBackward(boundaryNodes2.getSink(), sink);
			return new BoundaryNodesImpl(source, sink);
		}

		public Object visit(CompositionNCho node, Object data) throws CompositionVisitorException {
			Collection<BoundaryNodes> patterns = new ArrayList<>();
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				patterns.add((BoundaryNodes) node.jjtGetChild(i).jjtAccept(this, data));
			}
			return composeInNaryChoice(patterns);
		}

		private Object composeInNaryChoice(Collection<BoundaryNodes> patternBoundaries) {
			Place source = addAnonymousPlace();
			Place sink = addAnonymousPlace();
			for (BoundaryNodes boundary : patternBoundaries) {
				connectEliminateTauForward(source, boundary.getSource());
				connectEliminateTauBackward(boundary.getSink(), sink);
			}
			return new BoundaryNodesImpl(source, sink);
		}

		public Object visit(CompositionSeq node, Object data) throws CompositionVisitorException {
			BoundaryNodes boundaryNodes1 = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			BoundaryNodes boundaryNodes2 = (BoundaryNodes) node.jjtGetChild(1).jjtAccept(this, data);
			return composeInSequence(boundaryNodes1, boundaryNodes2);
		}

		private BoundaryNodes composeInSequence(BoundaryNodes boundaryNodes1, BoundaryNodes boundaryNodes2) {
			Place source = addAnonymousPlace();
			connectEliminateTauForward(source, boundaryNodes1.getSource());
			connectEliminateTauBackward(boundaryNodes1.getSink(), boundaryNodes2.getSource());
			Place sink = addAnonymousPlace();
			connectEliminateTauBackward(boundaryNodes2.getSink(), sink);
			return new BoundaryNodesImpl(source, sink);
		}

		public Object visit(CompositionInt node, Object data) throws CompositionVisitorException {
			Collection<BoundaryNodes> patterns = new ArrayList<>();
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				patterns.add((BoundaryNodes) node.jjtGetChild(i).jjtAccept(this, data));
			}
			return composeInInterleaving(patterns);
		}

		private BoundaryNodes composeInInterleaving(Collection<BoundaryNodes> patternBoundaries) {
			Place source = addAnonymousPlace();
			Place sink = addAnonymousPlace();
			Transition split = addInvisibleTransition();
			Transition merge = addInvisibleTransition();
			net.addArc(source, split);
			net.addArc(merge, sink);
			Place mutex = addAnonymousPlace();
			net.addArc(split, mutex);
			net.addArc(mutex, merge);
			for (BoundaryNodes boundary : patternBoundaries) {
				Place preStart = addAnonymousPlace();
				Transition start = addInvisibleTransition();
				// AND-split
				net.addArc(split, preStart);
				net.addArc(preStart, start);
				// Aquire mutex
				net.addArc(mutex, start);
				// Start
				net.addArc(start, boundary.getSource());

				Transition end = addInvisibleTransition();
				Place postEnd = addAnonymousPlace();
				net.addArc(boundary.getSink(), end);
				net.addArc(end, postEnd);
				// Release mutex
				net.addArc(end, mutex);
				// AND-join
				net.addArc(postEnd, merge);
			}
			return new BoundaryNodesImpl(source, sink);
		}

		public Object visit(CompositionRepInf node, Object data) throws CompositionVisitorException {
			BoundaryNodes child = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			addLoop(child);
			addSkip(child);
			return child;
		}

		public Object visit(CompositionRepInfOne node, Object data) throws CompositionVisitorException {
			BoundaryNodes child = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			addLoop(child);
			return child;
		}

		public Object visit(CompositionRepZeroOne node, Object data) throws CompositionVisitorException {
			BoundaryNodes child = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);
			addSkip(child);
			return child;
		}

		private void addSkip(BoundaryNodes child) {
			Transition skip = addInvisibleTransition();
			net.addArc(child.getSource(), skip);
			net.addArc(skip, child.getSink());
		}

		private void addLoop(BoundaryNodes child) {
			Transition redo = addInvisibleTransition();
			net.addArc(child.getSink(), redo);
			net.addArc(redo, child.getSource());
		}

		public Object visit(CompositionRepExact node, Object data) throws CompositionVisitorException {
			try {
				BoundaryNodes child = (BoundaryNodes) node.jjtGetChild(0).jjtAccept(this, data);

				long n = 0;
				long m = Integer.MAX_VALUE;

				if (node.getN().equals("0")) {
					addSkip(child);
				} else {
					n = Long.parseLong(node.getN());
				}
				if (node.getM().equals("*")) {
					addLoop(child);
				} else {
					m = Long.parseLong(node.getM());
				}

				Place source = addAnonymousPlace();
				Place sink = addAnonymousPlace();

				PNWDTransition initial = addInvisiblePNWDTransition();
				DataElement counter = net.addVariable(labelGenerator.getAnonymousVariableName(), Long.class, 0l, m + 1l);
				net.assignWriteOperation(initial, counter);
				initial.setGuard(net, counter.getVarName() + "' == 0");

				net.addArc(source, initial);
				Place pre = addAnonymousPlace();
				net.addArc(initial, pre);

				PNWDTransition canStop = addInvisiblePNWDTransition();
				net.addArc(pre, canStop);
				net.addArc(canStop, sink);

				net.assignReadOperation(canStop, counter);
				canStop.setGuard(net, counter.getVarName() + ">=" + n);

				PNWDTransition canExecute = addInvisiblePNWDTransition();
				net.addArc(pre, canExecute);
				net.addArc(canExecute, child.getSource());

				net.assignReadOperation(canExecute, counter);
				canExecute.setGuard(net, counter.getVarName() + "<" + m);

				PNWDTransition executed = addInvisiblePNWDTransition();
				net.addArc(child.getSink(), executed);
				net.addArc(executed, pre);

				net.assignWriteOperation(executed, counter);
				executed.setGuard(net, counter.getVarName() + "' == " + counter.getVarName() + "+1");

				net.assignReadOperation(executed, counter);

				return new BoundaryNodesImpl(source, sink);
			} catch (NumberFormatException | NonExistingVariableException
					| org.processmining.datapetrinets.expression.syntax.ParseException e) {
				throw new CompositionVisitorException("Invalid guard trying to compile exact repetition node", e);
			}
		}

		private PNWDTransition addInvisiblePNWDTransition() {
			return (PNWDTransition) addInvisibleTransition();
		}

		public Object visit(CompositionId node, Object data) throws CompositionVisitorException {
			AbstractionPattern pattern = patternMap.get(getLiteralValue(node));
			if (pattern != null) {
				DataPetriNet patternNet = pattern.getDPN();
				old2New.putAll(addDPNStructure(net, patternNet, labelGenerator));

				// Make sure to use the copied places
				Place oldSource = findSource(patternNet);
				Place oldSink = findSink(patternNet);
				if (oldSink != null && oldSource != null) {
					final Place source = (Place) old2New.get(oldSource);
					final Place sink = (Place) old2New.get(oldSink);
					return new BoundaryNodesImpl(source, sink);
				} else {
					throw new CompositionVisitorException("Invalid pattern " + getLiteralValue(node)
							+ " could not find a single sink/source node!");
				}
			} else {
				throw new CompositionVisitorException("Unkown pattern " + getLiteralValue(node));
			}
		}

		private String getLiteralValue(CompositionId node) {
			String id = (String) node.jjtGetValue();
			if (id.charAt(0) == '"') {
				id = id.substring(1, id.length() - 1);
			}
			return id;
		}

		private void connectEliminateTauForward(Place source, Place target) {
			Transition nextTransition = (Transition) getDirectSucessor(target);
			if (reduceTau && nextTransition != null) {
				// Single edge from source to this transition
				net.addArc(source, nextTransition);
				// Remove unnecessary target node
				net.removeNode(target);
			} else {
				// Introduce tau-transition
				Transition inv = addInvisibleTransition();
				net.addArc(source, inv);
				net.addArc(inv, target);
			}
		}

		private void connectEliminateTauBackward(Place source, Place target) {
			Transition prevTransition = (Transition) getDirectPredecessor(source);
			if (reduceTau && prevTransition != null) {
				// Single edge from prev to target
				net.addArc(prevTransition, target);
				// Remove unnecessary source node
				net.removeNode(source);
			} else {
				// Introduce tau-transition
				Transition inv = addInvisibleTransition();
				net.addArc(source, inv);
				net.addArc(inv, target);
			}
		}

		private PetrinetNode getDirectPredecessor(PetrinetNode to) {
			AbstractDirectedGraph<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> graph = to
					.getGraph();
			Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> iter = graph.getInEdges(to)
					.iterator();
			PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge = getSingleElement(iter);
			if (inEdge != null) {// Only a single in-edge
				return inEdge.getSource();
			} else {
				return null;
			}
		}

		private PetrinetNode getDirectSucessor(PetrinetNode from) {
			AbstractDirectedGraph<PetrinetNode, PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> graph = from
					.getGraph();
			Iterator<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>> iter = graph.getOutEdges(from)
					.iterator();
			PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> inEdge = getSingleElement(iter);
			if (inEdge != null) {// Only a single in-edge
				return inEdge.getTarget();
			} else {
				return null;
			}
		}

		private <T> T getSingleElement(Iterator<T> iter) {
			if (iter.hasNext()) {
				T pred = iter.next();
				if (!iter.hasNext()) {
					return pred;
				} else {
					return null;
				}
			} else {
				return null;
			}
		}

		private Transition addInvisibleTransition() {
			Transition transition = net.addTransition(labelGenerator.getAnonymousTransitionName());
			transition.setInvisible(true);
			return transition;
		}

		private Place addAnonymousPlace() {
			return net.addPlace(labelGenerator.getAnonymousPlaceName());
		}

		private static Place findSource(PetrinetGraph net) {
			for (Place p : net.getPlaces()) {
				if (Collections2.filter(net.getInEdges(p),
						new Predicate<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>>() {

							public boolean apply(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge) {
								return edge.getTarget() instanceof Place;
							}

						}).isEmpty()) {
					return p;
				}
			}
			return null;
		}

		private static Place findSink(PetrinetGraph net) {
			for (Place p : net.getPlaces()) {
				if (Collections2.filter(net.getOutEdges(p),
						new Predicate<PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode>>() {

							public boolean apply(PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge) {
								return edge.getSource() instanceof Place;
							}

						}).isEmpty()) {
					return p;
				}
			}
			return null;
		}

		private static Map<PetrinetNode, PetrinetNode> addDPNStructure(DataPetriNet targetModel,
				DataPetriNet sourceModel, LabelGenerator labelGenerator) {
			Map<PetrinetNode, PetrinetNode> old2New = new HashMap<PetrinetNode, PetrinetNode>();
			// First add all variable, we keep the same naming so callers need to be make sure variables do not overlap if not desired
			for (DataElement variable : sourceModel.getVariables()) {
				DataElement newVariable = targetModel.addVariable(variable.getVarName(), variable.getType(),
						variable.getMinValue(), variable.getMaxValue());
				old2New.put(variable, newVariable);
			}

			for (PetrinetEdge<? extends PetrinetNode, ? extends PetrinetNode> edge : sourceModel.getEdges()) {
				PetrinetNode newSourceNode = old2New.get(edge.getSource());
				if (newSourceNode == null) {
					newSourceNode = addNode(targetModel, edge.getSource(), labelGenerator);
					old2New.put(edge.getSource(), newSourceNode);
				}
				PetrinetNode newTargetNode = old2New.get(edge.getTarget());
				if (newTargetNode == null) {
					newTargetNode = addNode(targetModel, edge.getTarget(), labelGenerator);
					old2New.put(edge.getTarget(), newTargetNode);
				}
				addEdge(targetModel, newSourceNode, newTargetNode);
			}
			return old2New;
		}

		private static void addEdge(DataPetriNet model, PetrinetNode source, PetrinetNode target) {
			if (source instanceof Transition && target instanceof Place) {
				model.addArc((Transition) source, (Place) target);
			} else if (source instanceof Place && target instanceof Transition) {
				model.addArc((Place) source, (Transition) target);
			} else if (target instanceof DataElement) {
				model.assignWriteOperation((Transition) source, (DataElement) target);
			} else if (source instanceof DataElement) {
				model.assignReadOperation((Transition) source, (DataElement) target);
			} else {
				throw new IllegalArgumentException(
						String.format("Cannot add an edge between %s and %s", source, target));
			}
		}

		private static PetrinetNode addNode(DataPetriNet model, PetrinetNode node, LabelGenerator labelGenerator) {
			if (node instanceof Transition) {
				Transition transition = (Transition) node;
				Transition newTransition = model.addTransition(labelGenerator.getTransitionName(model, transition));
				newTransition.setInvisible(transition.isInvisible());
				if (transition instanceof PNWDTransition) {
					GuardExpression expression = ((PNWDTransition) transition).getGuardExpression();
					if (expression != null) {
						try {
							((PNWDTransition) newTransition).setGuard(model, expression);
						} catch (NonExistingVariableException e) {
							throw new RuntimeException(e);
						}
					}
				}
				return newTransition;
			} else if (node instanceof Place) {
				Place place = (Place) node;
				return model.addPlace(labelGenerator.getPlaceName(model, place));
			} else if (node instanceof DataElement) {
				throw new IllegalArgumentException(
						String.format("Variable %s should have been added beforehand!", node));
			} else {
				throw new IllegalArgumentException(String.format("Unkown node %s", node));
			}
		}

		public Map<PetrinetNode, PetrinetNode> getOld2NewMap() {
			return old2New;
		}

	}

	private PatternCompiler() {
		super();
	}

	public static AbstractionModel compile(CompositionRoot composition, String composedModelName,
			Map<String, AbstractionPattern> patternMap, boolean reduceTau) throws CompositionVisitorException {
		CompilingVisitor compilingVisitor = new CompilingVisitor(composedModelName, patternMap, reduceTau);
		DataPetriNetsWithMarkings dpn = (DataPetriNetsWithMarkings) compilingVisitor.visit(composition, null);
		AbstractionModel model = new AbstractionModel(composition, dpn, compilingVisitor.getOld2NewMap(),
				patternMap.values());
		return model;
	}

	public static AbstractionModel compile(String compositionFormula, String netName,
			Map<String, AbstractionPattern> patternMap, boolean reduceTau) throws ParseException,
			CompositionVisitorException {
		CompositionParser parser = new CompositionParser(compositionFormula);
		return compile(parser.parse(), netName, patternMap, reduceTau);
	}

}
