package org.processmining.logenhancement.abstraction.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.processmining.logenhancement.abstraction.model.syntax.CompositionCho;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionId;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionInt;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionNCho;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionNPar;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionPar;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParser;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParserTreeConstants;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionParserVisitor;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepExact;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepInf;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepInfOne;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRepZeroOne;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionRoot;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionSeq;
import org.processmining.logenhancement.abstraction.model.syntax.CompositionVisitorException;
import org.processmining.logenhancement.abstraction.model.syntax.Node;
import org.processmining.logenhancement.abstraction.model.syntax.ParseException;
import org.processmining.logenhancement.abstraction.model.syntax.SimpleNode;

public final class PatternSimplifier {

	private static final class SimplifyingVisitor implements CompositionParserVisitor {

		public Object visit(SimpleNode node, Object data) throws CompositionVisitorException {
			throw new IllegalStateException("No unamed nodes allowed!");
		}

		public Object visit(CompositionRoot node, Object data) throws CompositionVisitorException {
			if (node.jjtGetNumChildren() > 1) {
				throw new IllegalArgumentException("Invalid expression " + node
						+ " should not have been parsed! Top level element is only allowed to have one child.");
			}
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			CompositionRoot root = new CompositionRoot(CompositionParserTreeConstants.JJTROOT);
			root.jjtAddChild(child, 0);
			return root;
		}

		public Object visit(CompositionPar node, Object data) throws CompositionVisitorException {
			List<Node> children = new ArrayList<>();
			collectPar(node, children, data);
			if (children.size() == 2) {
				CompositionPar newNode = new CompositionPar(CompositionParserTreeConstants.JJTPAR);
				newNode.jjtAddChild(children.get(0), 0);
				newNode.jjtAddChild(children.get(1), 1);
				return newNode;
			} else {
				CompositionNPar newNode = new CompositionNPar(CompositionParserTreeConstants.JJTNPAR);
				int i = 0;
				for (Node newChild : children) {
					newNode.jjtAddChild(newChild, i++);
				}
				return newNode;
			}
		}

		private void collectPar(Node node, List<Node> children, Object data) throws CompositionVisitorException {
			Node child1 = node.jjtGetChild(0);
			Node child2 = node.jjtGetChild(1);
			if (child1 instanceof CompositionPar) {
				collectPar(child1, children, data);
			} else {
				children.add((Node)child1.jjtAccept(this, data));
			}
			if (child2 instanceof CompositionPar) {
				collectPar(child2, children, data);
			} else {
				children.add((Node)child2.jjtAccept(this, data));
			}
		}

		public Object visit(CompositionNPar node, Object data) throws CompositionVisitorException {
			Collection<Node> children = new ArrayList<>();
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				children.add((Node) node.jjtGetChild(i).jjtAccept(this, data));
			}
			CompositionNPar newNode = new CompositionNPar(CompositionParserTreeConstants.JJTNPAR);
			int i = 0;
			for (Node newChild : children) {
				newNode.jjtAddChild(newChild, i++);
			}
			return newNode;
		}

		public Object visit(CompositionCho node, Object data) throws CompositionVisitorException {
			Node child1 = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			Node child2 = (Node) node.jjtGetChild(1).jjtAccept(this, data);
			CompositionCho newNode = new CompositionCho(CompositionParserTreeConstants.JJTCHO);
			newNode.jjtAddChild(child1, 0);
			newNode.jjtAddChild(child2, 1);
			return newNode;
		}

		public Object visit(CompositionNCho node, Object data) throws CompositionVisitorException {
			Collection<Node> children = new ArrayList<>();
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				children.add((Node) node.jjtGetChild(i).jjtAccept(this, data));
			}
			CompositionNCho newNode = new CompositionNCho(CompositionParserTreeConstants.JJTNCHO);
			int i = 0;
			for (Node newChild : children) {
				newNode.jjtAddChild(newChild, i++);
			}
			return newNode;
		}

		public Object visit(CompositionSeq node, Object data) throws CompositionVisitorException {
			Node child1 = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			Node child2 = (Node) node.jjtGetChild(1).jjtAccept(this, data);
			CompositionSeq newNode = new CompositionSeq(CompositionParserTreeConstants.JJTSEQ);
			newNode.jjtAddChild(child1, 0);
			newNode.jjtAddChild(child2, 1);
			return newNode;
		}

		public Object visit(CompositionInt node, Object data) throws CompositionVisitorException {
			Collection<Node> children = new ArrayList<>();
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				children.add((Node) node.jjtGetChild(i).jjtAccept(this, data));
			}
			CompositionInt newNode = new CompositionInt(CompositionParserTreeConstants.JJTINT);
			int i = 0;
			for (Node newChild : children) {
				newNode.jjtAddChild(newChild, i++);
			}
			return newNode;
		}

		public Object visit(CompositionRepInf node, Object data) throws CompositionVisitorException {
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			if (isUnboundedInterleaving(child)) {
				child = rewriteUnboundedInterleaving((CompositionInt) child);
			}
			CompositionRepInf newPar = new CompositionRepInf(CompositionParserTreeConstants.JJTREPINF);
			newPar.jjtAddChild(child, 0);
			return newPar;
		}

		/**
		 * Rewrites an interleaving, in which every element can be skipped or
		 * repeated unbounded number of times into a nary-choice
		 * 
		 * @param node
		 *            an interleaving node
		 * @return a choice node
		 */
		private CompositionNCho rewriteUnboundedInterleaving(CompositionInt node) {
			// Interleaving is transformed to choice where every pattern can be skipped!
			CompositionNCho choiceNode = new CompositionNCho(CompositionParserTreeConstants.JJTNCHO);
			for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
				Node childNode = node.jjtGetChild(i);
				assert childNode instanceof CompositionRepInf
						|| childNode instanceof CompositionRepZeroOne : "Cannot transform interleaving to repeated choice! Not all patterns in the interleaving can be skipped!"; // precondition for this simplification
				assert childNode.jjtGetNumChildren() == 1 : "Repetition is to be unary!";
				choiceNode.jjtAddChild(childNode.jjtGetChild(0), i);
			}

			return choiceNode;
		}

		private boolean isUnboundedInterleaving(Node node) {
			if (node instanceof CompositionInt) {
				boolean allChildrenUnbounded = true;
				for (int i = 0; i < node.jjtGetNumChildren(); ++i) {
					Node childNode = node.jjtGetChild(i);
					allChildrenUnbounded &= childNode instanceof CompositionRepInf
							|| childNode instanceof CompositionRepZeroOne;
				}
				return allChildrenUnbounded;
			}
			return false;
		}

		public Object visit(CompositionRepInfOne node, Object data) throws CompositionVisitorException {
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			CompositionRepInfOne newNode = new CompositionRepInfOne(CompositionParserTreeConstants.JJTREPINFONE);
			newNode.jjtAddChild(child, 0);
			return newNode;
		}

		public Object visit(CompositionRepZeroOne node, Object data) throws CompositionVisitorException {
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			CompositionRepZeroOne newNode = new CompositionRepZeroOne(CompositionParserTreeConstants.JJTREPZEROONE);
			newNode.jjtAddChild(child, 0);
			return newNode;
		}

		public Object visit(CompositionRepExact node, Object data) throws CompositionVisitorException {
			Node child = (Node) node.jjtGetChild(0).jjtAccept(this, data);
			CompositionRepExact newNode = new CompositionRepExact(CompositionParserTreeConstants.JJTREPEXACT);
			newNode.jjtAddChild(child, 0);
			return newNode;
		}

		public Object visit(CompositionId node, Object data) throws CompositionVisitorException {
			return node; // No children
		}

	}

	private PatternSimplifier() {
		super();
	}

	public static CompositionRoot simplify(CompositionRoot composition) throws CompositionVisitorException {
		return (CompositionRoot) new SimplifyingVisitor().visit(composition, null);
	}

	public static CompositionRoot simplify(String compositionFormula)
			throws ParseException, CompositionVisitorException {
		CompositionParser composition = new CompositionParser(compositionFormula);
		return simplify(composition.parse());
	}

}