package org.processmining.logenhancement.enriching;

import java.util.Date;
import java.util.ListIterator;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeTimestamp;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;

public class AddElapsedTimePlugin {

	@Plugin(name = "Add Elapsed Time in Trace as Attribute to all Events", parameterLabels = { "A log" }, //
			help = "Adds a new attribute (continuous) with the elapsed time in the trace to all events.", categories = {
					PluginCategory.Enhancement }, returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void addElapsedTimeInTrace(UIPluginContext context, XLog log) throws UserCancelledException {

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		String elapsedTimeKey = ProMUIHelper.queryForString(context, "Specify name of the attribute to be created");
		TimeResolution resolution = ProMUIHelper.queryForObject(context,
				"Specify the resolution to be used for the elapsed time", TimeResolution.values());

		doAddElapsedTimeInTrace(progBar, log, elapsedTimeKey, resolution);
	}

	public void doAddElapsedTimeInTrace(Progress progress, XLog log, String elapsedTimeKey, TimeResolution resolution) {
		ListIterator<XTrace> traceIterator = log.listIterator();
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		while (traceIterator.hasNext()) {
			XTrace trace = traceIterator.next();

			long startTime = Long.MAX_VALUE;

			ListIterator<XEvent> eventIterator = trace.listIterator();
			while (eventIterator.hasNext()) {

				XEvent event = eventIterator.next();
				Date eventTime = XTimeExtension.instance().extractTimestamp(event);
				if (eventTime == null) {
					throw new RuntimeException("First event in trace is missing timestamp!");
				} else {
					startTime = Math.min(startTime, eventTime.getTime());
				}
			}

			ListIterator<XEvent> eventIter = trace.listIterator();
			while (eventIter.hasNext()) {
				XEvent event = eventIter.next();
				Date eventTime = XTimeExtension.instance().extractTimestamp(event);
				if (eventTime != null) {
					long elapsedTime = eventTime.getTime() - startTime;
					event.getAttributes().put(elapsedTimeKey, factory.createAttributeContinuous(elapsedTimeKey,
							(elapsedTime / (double) resolution.getFactor()), null));
				}
			}

			progress.inc();
		}
	}

	@Plugin(name = "Add Elapsed Time since Attribute until Event", parameterLabels = { "A log" }, //
			returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void addElapsedTimeSinceAttribute(UIPluginContext context, XLog log) throws UserCancelledException {

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		Boolean traceIncluded = ProMUIHelper.queryForObject(context, "Include trace attributes?",
				new Boolean[] { true, false });

		Set<String> timestampAttributes = new TreeSet<>();
		for (XTrace t : log) {
			if (traceIncluded) {
				for (XAttribute a : t.getAttributes().values()) {
					if (a instanceof XAttributeTimestamp) {
						timestampAttributes.add(a.getKey());
					}
				}
			}
			for (XEvent e : t) {
				for (XAttribute a : e.getAttributes().values()) {
					if (a instanceof XAttributeTimestamp) {
						timestampAttributes.add(a.getKey());
					}
				}
			}
		}

		String timeAttribute = ProMUIHelper.queryForString(context, "Select the attribute to use", timestampAttributes);

		String newAttributeName = ProMUIHelper.queryForString(context, "Specify name of the new attribute");
		TimeResolution resolution = ProMUIHelper.queryForObject(context,
				"Specify the resolution to be used for the elapsed time", TimeResolution.values());

		doAddElapsedTimeSinceAttribute(progBar, log, timeAttribute, newAttributeName, resolution);

	}

	public void doAddElapsedTimeSinceAttribute(Progress progressBar, XLog log, String timeAttributeKey,
			String newAttributeName, TimeResolution resolution) {

		ListIterator<XTrace> traceIterator = log.listIterator();
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		while (traceIterator.hasNext()) {
			XTrace trace = traceIterator.next();

			ListIterator<XEvent> eventIterator = trace.listIterator();
			while (eventIterator.hasNext()) {
				XEvent event = eventIterator.next();
				addElapsedTimeAttribute(timeAttributeKey, newAttributeName, resolution, factory, event);
			}

			progressBar.inc();
		}

	}

	private void addElapsedTimeAttribute(String timeAttributeKey, String newAttributeName, TimeResolution resolution,
			XFactory factory, XEvent event) {
		Date eventTime = XTimeExtension.instance().extractTimestamp(event);
		if (eventTime == null) {
			throw new RuntimeException("Event " + event + " is missing a timestamp!");
		}
		XAttribute timeAttribute = event.getAttributes().get(timeAttributeKey);
		if (timeAttribute != null) {
			if (!(timeAttribute instanceof XAttributeTimestamp)) {
				throw new RuntimeException(
						"Attribute " + timeAttribute + " is not of the expected type XTimestampAttribute!");
			}

			long attributeTime = ((XAttributeTimestamp) timeAttribute).getValueMillis();
			double elapsedTime = (eventTime.getTime() - attributeTime) / (double) resolution.getFactor();

			XAttributeContinuous elapsedTimeAttribute = factory.createAttributeContinuous(newAttributeName, elapsedTime,
					null);
			event.getAttributes().put(elapsedTimeAttribute.getKey(), elapsedTimeAttribute);
		}
	}

}
