package org.processmining.logenhancement.enriching;

import java.util.Date;
import java.util.ListIterator;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;

/**
 * Plugin that adds remaining time to all events on a trace.
 * Based on {@link AddElapsedTimePlugin} by F. Mannhardt
 * 
 * @author abolt
 *
 */
public class AddRemainingTimePlugin {
	
	@Plugin(name = "Add Remaining Time in Trace as Attribute to all Events", parameterLabels = { "A log" }, //
			help = "Adds a new attribute (continuous) with the remaining time in the trace to all events.", categories = {
					PluginCategory.Enhancement }, returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Alfredo Bolt", email = "a.bolt@tue.nl")
	public void addRemainingTimeInTrace(UIPluginContext context, XLog log) throws UserCancelledException {

		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		String elapsedTimeKey = ProMUIHelper.queryForString(context, "Specify name of the attribute to be created");
		TimeResolution resolution = ProMUIHelper.queryForObject(context,
				"Specify the resolution to be used for the remaining time", TimeResolution.values());

		doAddRemainingTimeInTrace(progBar, log, elapsedTimeKey, resolution);
	}

	public void doAddRemainingTimeInTrace(Progress progress, XLog log, String remainingTimeKey, TimeResolution resolution) {
		ListIterator<XTrace> traceIterator = log.listIterator();
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		while (traceIterator.hasNext()) {
			XTrace trace = traceIterator.next();

			long endTime = 0l;

			ListIterator<XEvent> eventIterator = trace.listIterator();
			while (eventIterator.hasNext()) {

				XEvent event = eventIterator.next();
				Date eventTime = XTimeExtension.instance().extractTimestamp(event);
				if (eventTime == null) {
					throw new RuntimeException("An event in trace is missing timestamp!");
				} else {
					endTime = Math.max(endTime, eventTime.getTime());
				}
			}

			ListIterator<XEvent> eventIter = trace.listIterator();
			while (eventIter.hasNext()) {
				XEvent event = eventIter.next();
				Date eventTime = XTimeExtension.instance().extractTimestamp(event);
				if (eventTime != null) {
					long remainingTime = endTime - eventTime.getTime();
					event.getAttributes().put(remainingTimeKey, factory.createAttributeContinuous(remainingTimeKey,
							(remainingTime / (double) resolution.getFactor()), null));
				}
			}

			progress.inc();
		}
	}
}
