package org.processmining.logenhancement.filtering;

import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.logenhancement.utils.LogEnhancementHelper;
import org.processmining.plugins.utils.ProvidedObjectHelper;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Ordering;

/**
 * Removes events from the log
 * 
 * @author F. Mannhardt
 *
 */
public class FilterEventsPlugin {

	@Plugin(name = "Filter events based on attribute values", parameterLabels = { "Event Log" }, //
			help = "Simple filter that removes events from traces based on attribute values.", categories = {
					PluginCategory.Filtering }, returnLabels = {
							"Filtered Log" }, returnTypes = { XLog.class }, userAccessible = true)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog filterEvents(UIPluginContext context, XLog log) throws UserCancelledException {

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);

		XEventClasses eventClassObj = logInfo.getEventClasses();
		List<XEventClass> eventClasses = ProMUIHelper.queryForObjects(context, "Select the considered event class(es)",
				Ordering.natural().sortedCopy(eventClassObj.getClasses()));

		Set<String> attributeNameSet = new TreeSet<>();
		for (XTrace t : log) {
			for (XEvent e : t) {
				if (containsSameEventClass(eventClassObj, eventClasses, e)) {
					attributeNameSet.addAll(e.getAttributes().keySet());
				}
			}
		}

		String attributeName = ProMUIHelper.queryForObject(context, "Select the attribute that should be filtered on",
				attributeNameSet);

		Set<XAttribute> valueSet = new TreeSet<>();
		for (XTrace t : log) {
			for (XEvent e : t) {
				if (containsSameEventClass(eventClassObj, eventClasses, e)) {
					XAttribute a = e.getAttributes().get(attributeName);
					if (a != null) {
						valueSet.add(a);
					}
				}
			}
		}

		Set<XAttribute> attributeValues = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
				"Select the attribute value(s) of the events that should be filtered out", valueSet));

		XLog newLog = (XLog) log.clone();

		doFilterEvents(newLog, eventClassObj, eventClasses, attributeName, attributeValues);

		context.getFutureResult(0).setLabel(
				ProvidedObjectHelper.getProvidedObjectLabel(context, log) + " - Filtered on " + attributeName);

		return newLog;
	}

	/**
	 * Filters the supplied log according to the supplied attribute values. Each
	 * event with a matching attribute value and event class is removed.
	 * 
	 * @param log
	 * @param eventClass
	 * @param consideredEventClasses
	 * @param attributeName
	 * @param attributeValues
	 * @param factory
	 */
	public void doFilterEvents(XLog log, XEventClasses eventClass, List<XEventClass> consideredEventClasses,
			String attributeName, Set<XAttribute> attributeValues) {

		for (XTrace t : log) {

			for (Iterator<XEvent> iterator = t.iterator(); iterator.hasNext();) {
				XEvent e = iterator.next();
				if (containsSameEventClass(eventClass, consideredEventClasses, e)) {
					XAttribute a = e.getAttributes().get(attributeName);
					if (a != null && attributeValues.contains(a)) {
						iterator.remove();
					}
				}
			}
		}

	}

	private boolean containsSameEventClass(XEventClasses eventClasses, List<XEventClass> targetEventClasses, XEvent e) {
		for (XEventClass eventClass : targetEventClasses) {
			if (isSameEventClass(eventClasses, eventClass, e)) {
				return true;
			}
		}
		return false;
	}

	private static boolean isSameEventClass(XEventClasses eventClasses, XEventClass eventClassB, XEvent e) {
		return eventClassB.equals(eventClasses.getClassOf(e));
	}

}