package org.processmining.logenhancement.filtering.attributes;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;

/**
 * Filter attributes of Events PLEASE NOTE: This filter will update the original
 * XLog instead of creating a new XLog.
 * 
 * @author F. Mannhardt
 * 
 */
@Plugin(name = "Filter/Edit Attributes of an Event Log (In Place)", parameterLabels = { "Event Log"}, 
		returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1,
		categories = { PluginCategory.Filtering },
		help = "Filters (i.e. removes) attributes by Event Name, Attribute Name or Attribute Name & Value. WARNING: Updates the supplied XLog for performance reasons!")
public class AttributeEditorPlugin {

	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	@PluginVariant(variantLabel = "Filter/Edit Attributes of an Event Log (In Place)", requiredParameterLabels = { 0 })
	public void filterLog(UIPluginContext context, XLog log) throws UserCancelledException {

		List<XEventClassifier> classifiers = log.getClassifiers();
		if (classifiers.isEmpty()) {
			classifiers = new ArrayList<>();
			classifiers.add(new XEventNameClassifier());
		}
		
		XEventClassifier classifier = ProMUIHelper.queryForObject(context, "Select Event Classifier", classifiers);
		LogAdapter logAdapter = new LogAdapter(log, classifier);
		
		AttributeEditorPanel uiPanel = new AttributeEditorPanel(logAdapter);
		if (uiPanel.getUserChoice(context) != InteractionResult.CANCEL) {

			List<AttributeEditorSelection> filterSelection = uiPanel.getFilterSelection();
			doFilterLog(context, logAdapter, filterSelection);

			DateFormat dateFormat = new SimpleDateFormat("HH:mm:ss");
			Date date = new Date();
			String timeString = dateFormat.format(date);

			String logName = XConceptExtension.instance().extractName(log);
			if (logName == null)
				logName = "log";
			logName = logName + " (filtered @ " + timeString + ")";
			context.getFutureResult(0).setLabel(logName);

		} else {
			context.log("Aborted by user!");
			context.getFutureResult(0).cancel(true);
		}

	}
	
	public XLog doFilterLog(PluginContext context, LogAdapter logAdapter, List<AttributeEditorSelection> filterSelection) {
		Progress progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(logAdapter.size() * filterSelection.size());
		progBar.setValue(0);
		
		//TODO improve speed by filtering for all filterSelections at once
		for (final AttributeEditorSelection selection : filterSelection) {
			context.log("Processing: " + selection.toString(), MessageLevel.NORMAL);
			AttributeFilterByEnum filterOption = selection.getFilterOption();
			String filterAttributeName = selection.getAttributeName();
			Object filterAttributeValue = selection.getAttributeValue();
			XEventClass filterEventClass = selection.getEventClass();
			XAttributeEditor attributeEditor = buildAttributeEditor(selection);
			new AttributeEditor(progBar, logAdapter, filterOption, filterAttributeName, filterAttributeValue,
					filterEventClass, attributeEditor).processLog();
		}
		
		return logAdapter.getLog();
	}

	private XAttributeEditor buildAttributeEditor(final AttributeEditorSelection selection) {
		return new XAttributeEditor() {

			public boolean shouldRemoveAttribute() {
				return selection.getReplacementValue().isEmpty();
			}

			public XAttribute editAttribute(XAttribute attribute) {
				if (!selection.getReplacementValue().isEmpty()) {
					if (attribute instanceof XAttributeLiteral) {
						return XFactoryRegistry.instance().currentDefault().createAttributeLiteral(attribute.getKey(), selection.getReplacementValue(), attribute.getExtension());
					} else if (attribute instanceof XAttributeDiscrete) {
						return XFactoryRegistry.instance().currentDefault().createAttributeDiscrete(attribute.getKey(), Integer.valueOf(selection.getReplacementValue()), attribute.getExtension());
					} else if (attribute instanceof XAttributeContinuous) {
						return XFactoryRegistry.instance().currentDefault().createAttributeContinuous(attribute.getKey(), Double.valueOf(selection.getReplacementValue()), attribute.getExtension());
					} else if (attribute instanceof XAttributeBoolean) {
						return XFactoryRegistry.instance().currentDefault().createAttributeBoolean(attribute.getKey(), Boolean.valueOf(selection.getReplacementValue()), attribute.getExtension());
					} else {
						throw new UnsupportedOperationException("Does not support manipulating "+attribute.getClass().getSimpleName());
					}					
				} else {
					return attribute;
				}
			}

			public void addAttributes(XAttributable a) {
				for (XAttribute attr: selection.getAdditionalAttributes()) {
					a.getAttributes().put(attr.getKey(), (XAttribute) attr.clone());
				}
			}
		};
	}

}
