package org.processmining.logenhancement.merge.alignmentbased;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NavigableSet;
import java.util.Set;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.plugins.balancedconformance.dataflow.datastate.DataStateFactory;
import org.processmining.plugins.balancedconformance.mapping.StringDiscretizer;

final class EventSetImpl implements EventSet {

	private final List<EventEntry> sortedEventEntries;
	private final BitSet usedEvents;
	private final Map<String, NavigableSet<EventEntry>> eventsByClass;
	private final XEventClassifier eventClasses;

	EventSetImpl(DataStateFactory stateFactoy, @SuppressWarnings("rawtypes") Map<String, Class> varType,
			Map<String, String> variableMapping, StringDiscretizer discretizer, XEventClassifier xEventClassifier,
			XLog[] logs) {
		super();
		this.eventClasses = xEventClassifier;
		ArrayList<XEvent> xEventList = new ArrayList<>();

		for (XLog log : logs) {
			for (XTrace t : log) {
				for (XEvent e : t) {
					xEventList.add(e);
				}
			}
		}
		Collections.sort(xEventList, new Comparator<XEvent>() {

			public int compare(XEvent o1, XEvent o2) {
				XTimeExtension instance = XTimeExtension.instance();
				return instance.extractTimestamp(o1).compareTo(XTimeExtension.instance().extractTimestamp(o2));
			}
		});

		this.usedEvents = new BitSet(xEventList.size());
		//TODO do i really need both the Map and the List?
		this.eventsByClass = new HashMap<>();
		this.sortedEventEntries = new ArrayList<>(xEventList.size());

		int eventCounter = 0;
		for (XEvent e : xEventList) {
			// Assign a index that in the correct order
			int eventIdx = eventCounter++;
			EventEntry eventEntry = new EventEntryImpl(e, eventIdx, stateFactoy, varType, variableMapping,
					discretizer);
			sortedEventEntries.add(eventEntry);
			registerEventClass(eventClasses.getClassIdentity(e), eventEntry);
		}

	}

	private void registerEventClass(String eventClassId, EventEntry eventEntry) {
		NavigableSet<EventEntry> eventSet = eventsByClass.get(eventClassId);
		if (eventSet == null) {
			eventSet = new TreeSet<>();
			eventsByClass.put(eventClassId, eventSet);
		}
		eventSet.add(eventEntry);
	}

	public void markUnavailable(EventEntry e) {
		usedEvents.set(e.getEventIndex());
	}

	public BitSet getUsedEvents() {
		return usedEvents;
	}

	public NavigableSet<EventEntry> findByClass(XEventClass neededEventClass) {
		return eventsByClass.get(neededEventClass.getId());
	}

	public Set<XEvent> asSet() {
		HashSet<XEvent> eventSet = new HashSet<>();
		for (EventEntry entry : sortedEventEntries) {
			eventSet.add(entry.getEvent());
		}
		return Collections.unmodifiableSet(eventSet);
	}

	public EventEntry getEvent(int index) {
		return sortedEventEntries.get(index);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.fm.project4.AlignmentBasedLogMerger.EventSet
	 * #getEvents(java.util.BitSet)
	 */
	public List<EventEntry> getEventsSortedByTime(BitSet eventBitSet) {
		final List<EventEntry> entryList = new ArrayList<>();
		for (int i = eventBitSet.nextSetBit(0); i >= 0; i = eventBitSet.nextSetBit(i + 1)) {
			entryList.add(sortedEventEntries.get(i));
		}
		return entryList;
	}

	public int size() {
		return sortedEventEntries.size();
	}
}