package org.processmining.logenhancement.transforming;

import java.security.SecureRandom;
import java.util.Date;

import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;

public class AnonymizeEventTimestampsPlugin {

	private enum TimeResolution {

		MILLISECONDS(1l, "Milliseconds"), // 
		SECONDS(1000l, "Seconds"), //
		MINUTES(1000l * 60l, "Minutes"), // 
		HOURS(1000l * 60l * 60l, "Hours"), //
		DAYS(1000l * 60l * 60l * 24l, "Days"), WEEKS(1000l * 60l * 60l * 24l * 7l, "Weeks");

		private final long factor;
		private final String unit;

		private TimeResolution(long millisecondFactor, String unit) {
			this.factor = millisecondFactor;
			this.unit = unit;
		}

		public long getFactor() {
			return factor;
		}

		public String getUnit() {
			return unit;
		}

		public String toString() {
			return getUnit();
		}

	}

	@Plugin(name = "Anonymize: Randomly shift time of traces/events", parameterLabels = { "A log" }, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the time:timestamp attributes of events in the log by moving the timestamps of events around randomly while keeping the distances within the trace.", // 
			returnLabels = { "Anonymized Log" }, returnTypes = { XLog.class })
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog anonymizeTimestamps(UIPluginContext context, XLog log) throws UserCancelledException {

		TimeResolution timeUnit = ProMUIHelper.queryForObject(context, "Choose time unit which is used to shift events",
				TimeResolution.values());
		int minimumShift = ProMUIHelper.queryForInteger(context,
				"Choose the minimum time (in " + timeUnit + ") with which events will be shifted");
		int maximumShift = ProMUIHelper.queryForInteger(context,
				"Choose the maximum time (in " + timeUnit + ") with which events will be shifted");

		return anonymizeTimestamps(context.getProgress(), (XLog) log.clone(), timeUnit, minimumShift, maximumShift);
	}

	@Plugin(name = "Anonymize: Randomly shift time of traces/events (in-place)", parameterLabels = { "A log" }, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the time:timestamp attributes of events in the log by moving the timestamps of events around randomly while keeping the distances within the trace.", // 
			returnLabels = {}, returnTypes = {}, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void anonymizeTimestampsInPlace(UIPluginContext context, XLog log) throws UserCancelledException {

		TimeResolution timeUnit = ProMUIHelper.queryForObject(context, "Choose time unit which is used to shift events",
				TimeResolution.values());
		int minimumShift = ProMUIHelper.queryForInteger(context,
				"Choose the minimum time (in " + timeUnit + ") with which events will be shifted");
		int maximumShift = ProMUIHelper.queryForInteger(context,
				"Choose the maximum time (in " + timeUnit + ") with which events will be shifted");

		anonymizeTimestamps(context.getProgress(), log, timeUnit, minimumShift, maximumShift);
	}

	public XLog anonymizeTimestamps(Progress progress, XLog log, TimeResolution timeUnit, int minimumShift,
			int maximumShift) {

		SecureRandom rnd = new SecureRandom();
		XTimeExtension timeExtension = XTimeExtension.instance();

		for (XTrace trace : log) {
			int nextShift = rnd.nextInt(maximumShift - minimumShift) + minimumShift;
			long nextMsShift = timeUnit.getFactor() * nextShift;
			if (rnd.nextBoolean()) {
				nextMsShift = -nextMsShift;
			}
			for (XEvent event : trace) {
				Date originalTime = timeExtension.extractTimestamp(event);
				timeExtension.assignTimestamp(event, originalTime.getTime() + nextMsShift);
			}
		}

		return log;
	}

}
