package org.processmining.logenhancement.transforming;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributable;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.base.Predicate;
import com.google.common.collect.BiMap;
import com.google.common.collect.HashBiMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.Sets;

public class AnonymizeLogPlugin {

	public enum AnonymizationType {

		ALL_EXCEPT_NAME("Remove all event attributes except classifier"), ALL_EXCEPT_TIME(
				"Remove all event attributes except classifier & time"), OBFUSCATE_EVENT(
						"Obfuscate event attribute values"), TRACE_EXCEPT_NAME(
								"Remove trace attributes except concept:name"), OBFUSCATE_TRACE(
										"Obfuscate trace attribute values"), OBFUSCATE_EVENT_ATTRIBUTE_NAMES(
												"Obfuscate event attribute names"), OBFUSCATE_TRACE_ATTRIBUTE_NAMES(
														"Obfuscate trace attribute names");

		private String desc;

		private AnonymizationType(String desc) {
			this.desc = desc;
		}

		@Override
		public String toString() {
			return desc;
		}

	}

	@Plugin(name = "Anonymize Event Log", parameterLabels = { "Event Log" }, returnLabels = {
			"Anonymized Event Log" }, returnTypes = { XLog.class }, userAccessible = true, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the events in an event log using different strategies. For example, removing attributes or obfuscating attributues.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog anonymise(final UIPluginContext context, XLog log) throws UserCancelledException {

		final AnonymizationType type = ProMUIHelper.queryForObject(context, "Choose Anonymization Strategy",
				AnonymizationType.values());

		return anonymiseWith(context, log, type);
	}

	@Plugin(name = "Anonymize: Obfuscate literal event attribute values", parameterLabels = {
			"Event Log" }, returnLabels = {
					"Anonymized Event Log" }, returnTypes = { XLog.class }, userAccessible = true, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the event attribtes in an event log by obfuscating literal values.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog obfuscateEventAttributes(final UIPluginContext context, XLog log) throws UserCancelledException {
		return anonymiseWith(context, log, AnonymizationType.OBFUSCATE_EVENT);
	}

	@Plugin(name = "Anonymize: Obfuscate event attribute names", parameterLabels = { "Event Log" }, returnLabels = {
			"Anonymized Event Log" }, returnTypes = { XLog.class }, userAccessible = true, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the trace attribtes in an event log by obfuscating their names.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog obfuscateEventAttributeNames(final UIPluginContext context, XLog log) throws UserCancelledException {
		return anonymiseWith(context, log, AnonymizationType.OBFUSCATE_EVENT_ATTRIBUTE_NAMES);
	}

	@Plugin(name = "Anonymize: Obfuscate literal trace attribute values", parameterLabels = {
			"Event Log" }, returnLabels = {
					"Anonymized Event Log" }, returnTypes = { XLog.class }, userAccessible = true, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the trace attributes in an event log by obfuscating literal values.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog obfuscateTraceAttributes(final UIPluginContext context, XLog log) throws UserCancelledException {
		return anonymiseWith(context, log, AnonymizationType.OBFUSCATE_TRACE);
	}

	@Plugin(name = "Anonymize: Obfuscate trace attribute names", parameterLabels = { "Event Log" }, returnLabels = {
			"Anonymized Event Log" }, returnTypes = { XLog.class }, userAccessible = true, //
			categories = {
					PluginCategory.Enhancement }, help = "Anonymizes the trace attributes in an event log by obfuscating their names.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog obfuscateTraceAttributeNames(final UIPluginContext context, XLog log) throws UserCancelledException {
		return anonymiseWith(context, log, AnonymizationType.OBFUSCATE_TRACE_ATTRIBUTE_NAMES);
	}

	private XLog anonymiseWith(final UIPluginContext context, XLog log, final AnonymizationType type)
			throws UserCancelledException {

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XFactory factory = XFactoryRegistry.instance().currentDefault();

		Set<String> attributesToAnonymize = new TreeSet<>();
		Map<String, String> obfuscationMapping = new HashMap<>();
		Map<String, Class<? extends XAttribute>> attributeType = new HashMap<>();

		if (type == AnonymizationType.OBFUSCATE_EVENT) {
			fillAttributeTablesWithEventAttributes(log, attributesToAnonymize, attributeType);
			attributesToAnonymize = ImmutableSet
					.copyOf(ProMUIHelper.queryForObjects(context, "Select event attributes that should be obfuscated",
							filterLiteralAttributes(attributesToAnonymize, attributeType)));
			obfuscationMapping = ProMUIHelper.queryMapToString(context,
					"Specify an (optional) prefix of the obfuscated value",
					filterLiteralAttributes(attributesToAnonymize, attributeType));
		} else if (type == AnonymizationType.OBFUSCATE_TRACE) {
			fillAttributeTablesWithTraceAttributes(log, attributesToAnonymize, attributeType);
			attributesToAnonymize = ImmutableSet
					.copyOf(ProMUIHelper.queryForObjects(context, "Select trace attributes that should be obfuscated",
							filterLiteralAttributes(attributesToAnonymize, attributeType)));
			obfuscationMapping = ProMUIHelper.queryMapToString(context,
					"Specify an (optional) prefix of the obfuscated value",
					filterLiteralAttributes(attributesToAnonymize, attributeType));
		} else if (type == AnonymizationType.OBFUSCATE_EVENT_ATTRIBUTE_NAMES) {
			fillAttributeTablesWithEventAttributes(log, attributesToAnonymize, attributeType);
			attributesToAnonymize = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
					"Select event attributes that should be obfuscated", attributesToAnonymize));
			obfuscationMapping = ProMUIHelper.queryMapToString(context, "Specify the obfuscated event attribute names",
					attributesToAnonymize, calcDefaultNameMapping(attributesToAnonymize));
		} else if (type == AnonymizationType.OBFUSCATE_TRACE_ATTRIBUTE_NAMES) {
			fillAttributeTablesWithTraceAttributes(log, attributesToAnonymize, attributeType);
			attributesToAnonymize = ImmutableSet.copyOf(ProMUIHelper.queryForObjects(context,
					"Select trace attributes that should be obfuscated", attributesToAnonymize));
			obfuscationMapping = ProMUIHelper.queryMapToString(context, "Specify the obfuscated trace attribute names",
					attributesToAnonymize, calcDefaultNameMapping(attributesToAnonymize));
		}

		XLog anonymizedLog = doAnonymizeLog(context.getProgress(), log, classifier, factory, attributesToAnonymize,
				attributeType, obfuscationMapping, type);
		return anonymizedLog;
	}

	private Map<String, String> calcDefaultNameMapping(Set<String> attributes) {
		Map<String, String> defaultMapping = new HashMap<>();
		int i = 0;
		for (String name : attributes) {
			defaultMapping.put(name, convertBase26(i++));
		}
		return defaultMapping;
	}

	private Set<String> filterLiteralAttributes(Set<String> attributesToAnonymize,
			final Map<String, Class<? extends XAttribute>> attributeType) {
		return Sets.filter(attributesToAnonymize, new Predicate<String>() {

			public boolean apply(String key) {
				return XAttributeLiteral.class.isAssignableFrom(attributeType.get(key));
			}
		});
	}

	private void fillAttributeTablesWithEventAttributes(XLog eventLog, Set<String> attributesToAnonymize,
			Map<String, Class<? extends XAttribute>> attributeType) {
		for (final XTrace trace : eventLog) {
			for (final XEvent event : trace) {
				fillAttributeTables(attributesToAnonymize, attributeType, event);
			}
		}
	}

	private void fillAttributeTablesWithTraceAttributes(XLog eventLog, Set<String> attributesToAnonymize,
			Map<String, Class<? extends XAttribute>> attributeType) {
		for (final XTrace trace : eventLog) {
			fillAttributeTables(attributesToAnonymize, attributeType, trace);
		}
	}

	private void fillAttributeTables(Set<String> attributesToAnonymize,
			Map<String, Class<? extends XAttribute>> attributeType, final XAttributable attributable) {
		for (XAttribute attribute : attributable.getAttributes().values()) {
			String key = attribute.getKey();
			Class<? extends XAttribute> clazz = attribute.getClass();
			attributeType.put(key, clazz);
			attributesToAnonymize.add(key);
		}
	}

	public XLog doAnonymizeLog(final Progress progress, XLog log, final XEventClassifier classifier,
			final XFactory factory, final Set<String> attributesToAnonymize,
			Map<String, Class<? extends XAttribute>> attributeType, final Map<String, String> obfuscationMap,
			final AnonymizationType type) {

		progress.setMinimum(0);
		progress.setMaximum(log.size());

		final Map<String, BiMap<String, String>> attributeMapping = new HashMap<>();
		// Calculate abstraction information for each attribute
		for (String attributeKey : attributesToAnonymize) {
			if (isLiteralAttribute(attributeType, attributeKey)) {
				// Nominal attribute -> Anonymization A,B,C,D,E, ...
				BiMap<String, String> anonymizationMapping = HashBiMap.create();
				attributeMapping.put(attributeKey, anonymizationMapping);
				int currentValue = 0;
				for (XTrace t : log) {
					if (type == AnonymizationType.OBFUSCATE_EVENT) {
						for (XEvent e : t) {
							currentValue = buildObfuscationTable(attributeKey, anonymizationMapping, currentValue, e);
						}
					} else if (type == AnonymizationType.OBFUSCATE_TRACE) {
						currentValue = buildObfuscationTable(attributeKey, anonymizationMapping, currentValue, t);
					}
				}
			}
		}
		
		System.out.println("Obfuscation coding:");
		for (Entry<String, BiMap<String, String>> attributeEntry: attributeMapping.entrySet()) {
			System.out.println(attributeEntry.getKey());
			Map<String, String> codeMap = new TreeMap<>(attributeEntry.getValue().inverse());
			for (Entry<String, String> mapping: codeMap.entrySet()) {
				System.out.println(mapping.getKey()+"="+mapping.getValue());	
			}
		}

		XLog anonymizedLog = factory.createLog(); // don't copy the metadata from the old log
		XUtils.assignConceptName(anonymizedLog, XUtils.getConceptName(log));

		for (XTrace trace : log) {

			XTrace anonymizedTrace = factory.createTrace();
			if (type == AnonymizationType.TRACE_EXCEPT_NAME) {
				XUtils.assignConceptName(anonymizedTrace, XUtils.getConceptName(trace));
			} else if (type == AnonymizationType.OBFUSCATE_TRACE) {
				obfuscateAttributes(trace, anonymizedTrace, attributeMapping, attributesToAnonymize, obfuscationMap,
						factory);
			} else if (type == AnonymizationType.OBFUSCATE_TRACE_ATTRIBUTE_NAMES) {
				obfuscateAttributeNames(obfuscationMap, trace, anonymizedTrace);
			} else {
				anonymizedTrace.getAttributes().putAll(trace.getAttributes());
			}

			for (XEvent event : trace) {
				XEvent anonymizedEvent = factory.createEvent();

				switch (type) {
					case ALL_EXCEPT_NAME :
						copyEventClassAttributes(classifier, event, anonymizedEvent);
						break;
					case ALL_EXCEPT_TIME :
						copyEventClassAttributes(classifier, event, anonymizedEvent);
						copyTime(event, anonymizedEvent);
						break;
					case OBFUSCATE_EVENT :
						obfuscateAttributes(event, anonymizedEvent, attributeMapping, attributesToAnonymize,
								obfuscationMap, factory);
						break;
					case OBFUSCATE_EVENT_ATTRIBUTE_NAMES :
						obfuscateAttributeNames(obfuscationMap, event, anonymizedEvent);
						break;
					case OBFUSCATE_TRACE :
					case OBFUSCATE_TRACE_ATTRIBUTE_NAMES :
					case TRACE_EXCEPT_NAME :
					default :
						anonymizedEvent.getAttributes().putAll((XAttributeMap) event.getAttributes().clone());
						break;
				}
				anonymizedTrace.add(anonymizedEvent);
			}
			progress.inc();
			anonymizedLog.add(anonymizedTrace);
		}

		return anonymizedLog;

	}

	private void obfuscateAttributeNames(final Map<String, String> obfuscationMap, XAttributable attributable,
			XAttributable anonymizedAttributable) {
		for (XAttribute attribute : attributable.getAttributes().values()) {
			String newKey = obfuscationMap.get(attribute.getKey());
			if (newKey != null) {
				XAttribute obfuscatedAttribute = XUtils.cloneAttributeWithChangedKey(attribute, newKey);
				XUtils.putAttribute(anonymizedAttributable, obfuscatedAttribute);
			} else {
				XUtils.putAttribute(anonymizedAttributable, (XAttribute) attribute.clone());
			}
		}
	}

	private int buildObfuscationTable(String attributeKey, Map<String, String> anonymizationMapping, int currentValue,
			XAttributable a) {
		XAttribute attribute = a.getAttributes().get(attributeKey);
		if (attribute != null) {
			String realValue = ((XAttributeLiteral) attribute).getValue();
			String anonValue = anonymizationMapping.get(realValue);
			if (anonValue == null) {
				anonValue = convertBase26(currentValue);
				anonymizationMapping.put(realValue, anonValue);
				return currentValue + 1;
			}
		}
		return currentValue;
	}

	private boolean isLiteralAttribute(Map<String, Class<? extends XAttribute>> attributeType, String attributeKey) {
		return XAttributeLiteral.class.isAssignableFrom(attributeType.get(attributeKey));
	}

	private void obfuscateAttributes(XAttributable attributable, XAttributable obfuscatedAttributable,
			Map<String, BiMap<String, String>> literalAttributeMapping, Set<String> attributesToAnonymize,
			Map<String, String> anonValuePrefix, XFactory factory) {

		for (XAttribute attr : attributable.getAttributes().values()) {
			String attrKey = attr.getKey();
			if (attributesToAnonymize.contains(attrKey)) {
				Map<String, String> literalMapping = literalAttributeMapping.get(attrKey);
				if (literalMapping != null) {
					String anonValue = literalMapping.get(((XAttributeLiteral) attr).getValue());
					String prefix = anonValuePrefix.get(attrKey);
					if (prefix != null) {
						anonValue = prefix.concat(anonValue);
					}
					obfuscatedAttributable.getAttributes().put(attrKey,
							factory.createAttributeLiteral(attrKey, anonValue, attr.getExtension()));
				}
				// Otherwise, remove as it cannot be easily anonymized
			} else {
				obfuscatedAttributable.getAttributes().put(attrKey, attr);
			}
		}

	}

	private void copyEventClassAttributes(final XEventClassifier classifier, XEvent event, XEvent newEvent) {
		String[] classifierAttributes = classifier.getDefiningAttributeKeys();
		for (String attributeKey : classifierAttributes) {
			newEvent.getAttributes().put(attributeKey, event.getAttributes().get(attributeKey));
		}
	}

	private void copyTime(XEvent event, XEvent newEvent) {
		Date oldTime = XTimeExtension.instance().extractTimestamp(event);
		if (oldTime != null) {
			XTimeExtension.instance().assignTimestamp(newEvent, oldTime);
		}
	}

	private static String convertBase26(int value) {
		value++;
		StringBuilder sb = new StringBuilder();
		while (value > 0) {
			value--;
			int remainder = value % 26;
			char alpha = (char) (remainder + 65);
			sb.append(alpha);
			value = (value - remainder) / 26;
		}

		return sb.toString();
	}

}
