package org.processmining.logenhancement.transforming;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeBoolean;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeLiteral;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Sets;

public class ChangeAttributeTypePlugin {

	@Plugin(name = "Change attribute type (In place)", parameterLabels = { "A log" }, //
			help = "Changes the type of an attribute. The existing value will be converted using the Java 'parse' methods (e.g., Boolean.parseBoolean, Long.parseLong, etc.).", categories = {
					PluginCategory.Enhancement }, returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void changeAttributeType(UIPluginContext context, XLog log) throws UserCancelledException {

		context.getProgress().setMaximum(log.size());

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);

		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);
		Set<XEventClass> eventNames = Sets.newTreeSet(logInfo.getEventClasses().getClasses());

		List<XEventClass> selectedEventClasses = ProMUIHelper.queryForObjects(context,
				"Please select the event classes for which the attribute type should be changed", eventNames);

		List<String> attributeKeys = new ArrayList<>(logInfo.getEventAttributeInfo().getAttributeKeys());
		Collections.sort(attributeKeys);

		String attributeName = ProMUIHelper.queryForString(context,
				"Please select the attribute that should be changed", attributeKeys);

		Class<? extends XAttribute> changedAttributeType = ProMUIHelper.queryForObject(context,
				"Please provide the new type of the attribute", ImmutableList.of(XAttributeLiteral.class,
						XAttributeDiscrete.class, XAttributeContinuous.class, XAttributeBoolean.class));

		Boolean failOnError = ProMUIHelper.queryForObject(context, "Fail on non-convertible attributes?",
				new Boolean[] { Boolean.TRUE, Boolean.FALSE });

		XFactory factory = XFactoryRegistry.instance().currentDefault();

		doChangeAttributeType(factory, log, logInfo, selectedEventClasses, attributeName, changedAttributeType,
				failOnError);

	}

	public void doChangeAttributeType(XFactory factory, XLog log, XLogInfo logInfo,
			List<XEventClass> selectedEventClasses, String attributeName,
			Class<? extends XAttribute> changedAttributeType, boolean failOnError) {
		for (XTrace t : log) {
			for (XEvent e : t) {
				if (selectedEventClasses.contains(logInfo.getEventClasses().getClassOf(e))) {
					XAttribute attribute = e.getAttributes().get(attributeName);
					if (attribute != null) {
						try {
							if (changedAttributeType == XAttributeLiteral.class) {
								XUtils.putAttribute(e,
										factory.createAttributeLiteral(attributeName, attribute.toString(), null));
							} else if (changedAttributeType == XAttributeDiscrete.class) {
								XUtils.putAttribute(e, factory.createAttributeDiscrete(attributeName,
										Long.parseLong(attribute.toString()), null));
							} else if (changedAttributeType == XAttributeContinuous.class) {
								XUtils.putAttribute(e, factory.createAttributeContinuous(attributeName,
										Double.parseDouble(attribute.toString()), null));
							} else if (changedAttributeType == XAttributeBoolean.class) {
								XUtils.putAttribute(e, factory.createAttributeBoolean(attributeName,
										Boolean.parseBoolean(attribute.toString()), null));
							} else {
								throw new RuntimeException("Unkown type " + changedAttributeType);
							}
						} catch (NumberFormatException e1) {
							if (failOnError) {
								throw e1;
							} // otherwise ignore
						}
					}
				}
			}
		}
	}

}
