package org.processmining.logenhancement.transforming;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

public class CopyEventAttributePlugin {

	public enum CopyType {
		FIRST, ALL, LAST
	}

	@Plugin(name = "Copy attributes between events (In place)", parameterLabels = { "A log" }, //
			categories = {
					PluginCategory.Enhancement }, help = "Copies attribute from a source event class to target event classes.", returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void copyAttribute(UIPluginContext context, XLog log) throws UserCancelledException {

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);

		XEventClasses eventClassObj = logInfo.getEventClasses();
		Collection<XEventClass> eventClasses = eventClassObj.getClasses();

		XEventClass sourceEvent = ProMUIHelper.queryForObject(context,
				"Select the source event class (1st event of this class will be used)", eventClasses);
		List<XEventClass> targetEvents = ProMUIHelper.queryForObjects(context, "Select the target event class(es)",
				eventClasses);

		List<String> attributeKeys = new ArrayList<>(logInfo.getEventAttributeInfo().getAttributeKeys());
		Collections.sort(attributeKeys);
		String attributeName = ProMUIHelper.queryForObject(context, "Select the attribute that should be copied",
				attributeKeys);
		String newAttributeName = ProMUIHelper.queryForString(context, "Specify 'name' of new attribute", null,
				attributeName);

		CopyType type = ProMUIHelper.queryForObject(context, "Specify target type of the copy operation",
				CopyType.values());

		doCopyAttribute(log, eventClassObj, sourceEvent, targetEvents, attributeName, newAttributeName, type);

	}

	public void doCopyAttribute(XLog log, XEventClasses eventClassObj, XEventClass sourceEvent,
			List<XEventClass> targetEventClasses, String attributeName, String newAttributeName, CopyType type) {
		for (XTrace xTrace : log) {

			XAttribute sourceAttribute = null;
			for (XEvent e : xTrace) {
				if (isSameEventClass(eventClassObj, sourceEvent, e)) {
					sourceAttribute = e.getAttributes().get(attributeName);
					// Just take the first occurence
					break;
				}
			}

			XEvent targetEvent = null;
			if (sourceAttribute != null) {
				for (XEvent e : xTrace) {
					if (containsSameEventClass(eventClassObj, targetEventClasses, e)
							&& (targetEvent == null || type == CopyType.ALL)) {
						e.getAttributes().put(newAttributeName,
								XUtils.cloneAttributeWithChangedKey(sourceAttribute, newAttributeName));
						targetEvent = e;
					} else if (containsSameEventClass(eventClassObj, targetEventClasses, e)) {
						targetEvent = e;
					}
				}
				if (targetEvent != null && type == CopyType.LAST) {
					targetEvent.getAttributes().put(newAttributeName,
							XUtils.cloneAttributeWithChangedKey(sourceAttribute, newAttributeName));

				}
			}
		}
	}

	private boolean containsSameEventClass(XEventClasses eventClasses, List<XEventClass> targetEventClasses, XEvent e) {
		for (XEventClass eventClass : targetEventClasses) {
			if (isSameEventClass(eventClasses, eventClass, e)) {
				return true;
			}
		}
		return false;
	}

	private static boolean isSameEventClass(XEventClasses eventClasses, XEventClass eventClassB, XEvent e) {
		return eventClassB.equals(eventClasses.getClassOf(e));
	}

}
