package org.processmining.logenhancement.transforming;

import java.util.Date;
import java.util.HashSet;
import java.util.ListIterator;
import java.util.Objects;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XLifecycleExtension;
import org.deckfour.xes.extension.std.XTimeExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

import com.google.common.collect.Ordering;

public class MergeSubsequentEvents {

	private static final int TIME_DIFFERENCE = 1000 * 60;

	public enum MergeFilter {
		SAME_CLASS("Compare event class") {
			public boolean sameEvent(XEventClassifier classifier, Set<String> relevantAttributes, XEvent eventA,
					XEvent eventB) {
				return classifier.sameEventClass(eventA, eventB);
			}
		},
		SHORT_TIME("Compare event timestamps") {
			public boolean sameEvent(XEventClassifier classifier, Set<String> relevantAttributes, XEvent eventA,
					XEvent eventB) {
				Date timeA = XTimeExtension.instance().extractTimestamp(eventA);
				Date timeB = XTimeExtension.instance().extractTimestamp(eventB);
				if (timeA != null && timeB != null) {
					return Math.abs(timeB.getTime() - timeA.getTime()) < TIME_DIFFERENCE;
				} else {
					return false;
				}
			}
		},
		SAME_DATA("Compare event class & attributes") {
			public boolean sameEvent(XEventClassifier classifier, Set<String> relevantAttributes, XEvent eventA,
					XEvent eventB) {
				boolean sameClass = classifier.sameEventClass(eventA, eventB);
				if (sameClass) {
					XAttributeMap attributesA = eventA.getAttributes();
					XAttributeMap attributesB = eventB.getAttributes();
					for (String key : relevantAttributes) {
						XAttribute attributeA = attributesA.get(key);
						XAttribute attributeB = attributesB.get(key);
						if (!Objects.equals(attributeA, attributeB)) {
							return false;
						}
					}
					return true;
				}
				return false;
			}
		};

		private final String label;

		private MergeFilter(String label) {
			this.label = label;
		}

		public String toString() {
			return label;
		}

		public abstract boolean sameEvent(XEventClassifier classifier, Set<String> relevantAttributes,
				XEvent subsequentEvent, XEvent currentEvent);
	}

	public enum MergeType {
		PLAIN_FIRST("Merge taking first event") {
			public long merge(XEventClasses eventClasses, XFactory factory, long instance, XTrace newTrace,
					ListIterator<XEvent> iterator, XEvent currentEvent, MergeFilter filter,
					Set<String> relevantAttributes) {
				// Just add first
				newTrace.add(factory.createEvent((XAttributeMap) currentEvent.getAttributes().clone()));

				// Look forward for more event classes of the same type
				while (iterator.hasNext()) {
					XEvent subsequentEvent = iterator.next();
					if (!filter.sameEvent(eventClasses.getClassifier(), relevantAttributes, subsequentEvent,
							currentEvent)) {
						// Go back and stop merging
						iterator.previous();
						break;
					}
				}
				return ++instance;
			}
		},
		PLAIN_LAST("Merge taking last event") {
			public long merge(XEventClasses eventClasses, XFactory factory, long instance, XTrace newTrace,
					ListIterator<XEvent> iterator, XEvent currentEvent, MergeFilter filter,
					Set<String> relevantAttributes) {
				XEvent lastEvent = currentEvent;

				// Look forward for more event classes of the same type
				while (iterator.hasNext()) {
					lastEvent = iterator.next();
					if (!filter.sameEvent(eventClasses.getClassifier(), relevantAttributes, lastEvent, currentEvent)) {
						// Go back and stop merging
						iterator.previous();
						// Go back to the last event with the same event class
						lastEvent = iterator.previous();
						iterator.next();
						break;
					}
				}

				// Add last event
				newTrace.add(factory.createEvent((XAttributeMap) lastEvent.getAttributes().clone()));
				return ++instance;
			}
		},
		LIFECYCLE("Merge take first as 'start' and last as 'complete' life-cycle transitions") {

			public long merge(XEventClasses eventClasses, XFactory factory, long instance, XTrace newTrace,
					ListIterator<XEvent> iterator, XEvent currentEvent, MergeFilter filter,
					Set<String> relevantAttributes) {
				XEvent newStartEvent = factory.createEvent((XAttributeMap) currentEvent.getAttributes().clone());
				XLifecycleExtension.instance().assignStandardTransition(newStartEvent,
						XLifecycleExtension.StandardModel.START);
				XConceptExtension.instance().assignInstance(newStartEvent, String.valueOf(instance));

				newTrace.add(newStartEvent);

				XEvent lastEvent = currentEvent;

				// Look forward for more event classes of the same type
				while (iterator.hasNext()) {
					lastEvent = iterator.next();
					if (!filter.sameEvent(eventClasses.getClassifier(), relevantAttributes, lastEvent, currentEvent)) {
						// Stop merging, go back so this event is being picked-up by the main iteration
						iterator.previous();
						// Go back once again and remember last time stamp
						lastEvent = iterator.previous();
						// Go forward again :)
						iterator.next();
						break;
					}
				}

				// Copy last found event
				lastEvent = factory.createEvent((XAttributeMap) lastEvent.getAttributes().clone());

				XLifecycleExtension.instance().assignStandardTransition(lastEvent,
						XLifecycleExtension.StandardModel.COMPLETE);
				XConceptExtension.instance().assignInstance(lastEvent, String.valueOf(instance));
				newTrace.add(lastEvent);

				return ++instance;
			}

		};

		private final String label;

		private MergeType(String label) {
			this.label = label;
		}

		public String toString() {
			return label;
		}

		public abstract long merge(XEventClasses eventClasses, XFactory factory, long instance, XTrace newTrace,
				ListIterator<XEvent> iterator, XEvent currentEvent, MergeFilter mergeFilter,
				Set<String> relevantAttributes);

	}

	@Plugin(name = "Merge subsequent events (AAABB -> AB)", parameterLabels = {}, //
			returnLabels = { "Merged Log" }, returnTypes = {
					XLog.class }, userAccessible = true, mostSignificantResult = 1, //
			categories = { PluginCategory.Enhancement }, keywords = { "subsequent", "merge", "duplicate", "collapse",
					"repeated" }, //
			help = "Merges events of the same event class directly following each other (subsequent). "
					+ "Remove repeated / duplicated events. For example a trace AAABBCAA would be transformed to ABCA.")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public XLog mergeSubsequentEvents(UIPluginContext context, XLog log) throws UserCancelledException {
		final Progress progress = context.getProgress();
		progress.setMaximum(log.size());

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XEventClasses eventClasses = XUtils.createEventClasses(classifier, log);
		XFactory f = XFactoryRegistry.instance().currentDefault();

		Set<XEventClass> relevantClasses = new HashSet<>(
				ProMUIHelper.queryForObjects(context, "Select which event classes should be considered",
						Ordering.natural().sortedCopy(eventClasses.getClasses())));

		MergeFilter mergeFilter = ProMUIHelper.queryForObject(context, "Select how to compare events",
				MergeFilter.values());

		Set<String> relevantAttributes = new HashSet<String>();
		if (mergeFilter == MergeFilter.SAME_DATA) {
			relevantAttributes = new HashSet<>(ProMUIHelper.queryForStrings(context,
					"Select which attributes should be considered in the comparison",
					Ordering.natural().sortedCopy(XUtils.getEventAttributeKeys(log))));
		}

		MergeType mergeType = ProMUIHelper.queryForObject(context, "Select merge type", MergeType.values());

		return doMergeSubsequentEvents(log, eventClasses, f, relevantClasses, relevantAttributes, mergeFilter,
				mergeType);
	}

	public XLog doMergeSubsequentEvents(XLog log, XEventClasses eventClasses, XFactory factory,
			Set<XEventClass> consideredClasses, Set<String> relevantAttributes, MergeFilter mergeFilter,
			MergeType mergeType) {
		XLog newLog = XUtils.createLogFrom(log, factory);
		long instance = 0;

		for (XTrace t : log) {
			XTrace newTrace = factory.createTrace((XAttributeMap) t.getAttributes().clone());

			for (ListIterator<XEvent> iterator = t.listIterator(); iterator.hasNext();) {
				XEvent currentEvent = iterator.next();
				XEventClass eventClass = eventClasses.getClassOf(currentEvent);

				// Check for subsequent events with same class
				if (consideredClasses.contains(eventClass)) {

					instance = mergeType.merge(eventClasses, factory, instance, newTrace, iterator, currentEvent,
							mergeFilter, relevantAttributes);

				} else {
					// Just copy existing
					newTrace.add(factory.createEvent((XAttributeMap) currentEvent.getAttributes().clone()));
				}

			}
			newLog.add(newTrace);
		}

		return newLog;
	}
}