package org.processmining.logenhancement.transforming;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.util.ui.widgets.helper.ProMUIHelper;
import org.processmining.framework.util.ui.widgets.helper.UserCancelledException;
import org.processmining.log.utils.XUtils;
import org.processmining.logenhancement.utils.LogEnhancementHelper;

/**
 * Assigns attributes are simply named "value" a more reasonable name. To be
 * used, for example, when there is an attribute named "value" on every Event
 * that means something different.
 * 
 * @author F. Mannhardt
 * 
 */
public class TransformValueAttributePlugin {

	private Progress progBar;

	@Plugin(name = "Transform 'value' attributes based on event class (In place)", parameterLabels = { "Event Log" }, //
	returnLabels = {}, returnTypes = {}, userAccessible = true, mostSignificantResult = -1, categories = { PluginCategory.Enhancement },
	help = "Sometimes attributes in an event log are simply named 'value', but their semantics depends on which event recorded the 'value'. "
			+ "This plug-in assigns a those attribute a better name derived from the event class. ")
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public void transformValueAttribute(UIPluginContext context, XLog log) throws UserCancelledException {

		progBar = context.getProgress();
		progBar.setMinimum(0);
		progBar.setMaximum(log.size());
		progBar.setValue(0);

		XEventClassifier classifier = LogEnhancementHelper.queryEventClassifier(context, log);
		XLogInfo logInfo = XLogInfoFactory.createLogInfo(log, classifier);
		XEventClasses eventClassesObj = logInfo.getEventClasses();

		String valueAttributeName = ProMUIHelper.queryForObject(context, "Select 'value' attribute", logInfo
				.getEventAttributeInfo().getAttributeKeys());

		Set<XEventClass> classesWithAttribute = new HashSet<>();

		for (XTrace t : log) {
			for (XEvent e : t) {
				if (e.getAttributes().containsKey(valueAttributeName)) {
					classesWithAttribute.add(eventClassesObj.getClassOf(e));
				}
			}
		}

		Map<XEventClass, String> mapping = ProMUIHelper.queryMapToString(context, "Configure the new names of the 'value' attributes",
				classesWithAttribute);
		
		for (XTrace t : log) {
			for (XEvent e : t) {
				XAttribute attribute = e.getAttributes().get(valueAttributeName);
				if (attribute != null) {
					XEventClass eventClass = eventClassesObj.getClassOf(e);
					String targetName = mapping.get(eventClass);
					if (targetName != null) {
						e.getAttributes().remove(valueAttributeName);
						e.getAttributes().put(targetName, XUtils.cloneAttributeWithChangedKey(attribute, targetName));						
					}
				}
			}
			progBar.inc();
		}

	}

}
