package org.processmining.logenhancement.transforming.classifier;

import java.util.HashMap;
import java.util.Map;

import org.deckfour.xes.classification.XEventAndClassifier;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.classification.XEventLifeTransClassifier;
import org.deckfour.xes.classification.XEventNameClassifier;

/**
 * Parameters for the transformation of an event log.
 * 
 * @author Eduardo González López de Murillas
 * 
 */
public class ActivityMappingParameters {

	/**
	 * Classifier parameter. This determines which classifier will be used
	 * during the transformation.
	 */
	private XEventClassifier classifier;
	private Map<String,String> activityMap;
	private String newClassifierName;

	/**
	 * Create default parameter values.
	 */
	public ActivityMappingParameters() {
		classifier = new XEventAndClassifier(new XEventNameClassifier(), new XEventLifeTransClassifier());
		activityMap = new HashMap<>();
		newClassifierName = "New Activity Mapping Classifier";
	}

	/**
	 * Set the classifier to the given classifier.
	 * 
	 * @param classifier
	 *            The given classifier.
	 */
	public void setClassifier(XEventClassifier classifier) {
		if (classifier != null) {
			this.classifier = classifier;
		}
	}

	/**
	 * Gets the classifier.
	 * 
	 * @return The classifier.
	 */
	public XEventClassifier getClassifier() {
		return classifier;
	}
	
	/**
	 * Set the new classifier name.
	 * 
	 * @param classifierName
	 *            The given classifier name.
	 */
	public void setNewClassifierName(String classifierName) {
		if (classifierName != null) {
			this.newClassifierName = classifierName;
		}
	}

	/**
	 * Gets the name of the new classifier.
	 * 
	 * @return The name of the classifier.
	 */
	public String getNewClassifierName() {
		return newClassifierName;
	}
	
	/**
	 * Set the new Activity Mapping.
	 * 
	 * @param activityMap
	 *            The given Activity Mapping.
	 */
	public void setActivityMapping(Map<String,String> activityMap) {
		if (activityMap != null) {
			this.activityMap = activityMap;
		}
	}

	/**
	 * Gets the Activity Mapping.
	 * 
	 * @return The Activity Mapping.
	 */
	public Map<String,String> getActivityMapping() {
		return activityMap;
	}

	/**
	 * Returns whether these parameter values are equal to the given parameter
	 * values.
	 * 
	 * @param object
	 *            The given parameter values.
	 * @return Whether these parameter values are equal to the given parameter
	 *         values.
	 */
	public boolean equals(Object object) {
		if (object instanceof ActivityMappingParameters) {
			ActivityMappingParameters parameters = (ActivityMappingParameters) object;
			if (classifier.equals(parameters.classifier)) {
				if (newClassifierName.equals(parameters.newClassifierName)) {
					if (activityMap.equals(parameters.activityMap)) {
						return true;
					}
				}
			}
		}
		return false;
	}
	
	/**
	 * Returns the hash code for these parameters.
	 */
	public int hashCode() {
		String code = classifier.hashCode()+"#"+newClassifierName.hashCode()+"#"+activityMap.hashCode();
		return code.hashCode();
	}
}
