package org.processmining.logenhancement.transforming.classifier;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventAttributeClassifier;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeLiteralImpl;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.Progress;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;

/**
 * Plug-in that transforms a log at attribute level.
 * 
 * @author Eduardo Gonzalez Lopez de Murillas
 * 
 */
public class ActivityMappingPlugin {

	/**
	 * Transform using provided parameter values.
	 * 
	 * @param context
	 *            The given plug-in context.
	 * @param log
	 *            The given event log.
	 * @param parameters
	 *            The given parameters to select the right classifier
	 * @return The transformed event log with added attributes and classifier.
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Eduardo Gonzalez Lopez de Murillas", email = "e.gonzalez@tue.nl")
	@Plugin(name = "Add Classifier based on Mapping of Activity Names (parameters)", returnLabels = {
			"Transformed Log" }, returnTypes = { XLog.class }, parameterLabels = { "Log",
					"Activity Mapping Parameters" }, userAccessible = true, categories = {
							PluginCategory.Enhancement }, help = "Adds a new classifier to the event log based on a user-configureable mapping from activity names of an existing classifier to the new classifier.")
	public XLog mapActivityNameDefault(PluginContext context, XLog log, ActivityMappingParameters parameters) {
		XLog newLog = (XLog) log.clone();
		return addActivityMappingName(context.getProgress(), newLog, parameters);
	}

	/**
	 * Transform using provided parameter values.
	 * 
	 * @param context
	 *            The given GUI-aware plug-in context.
	 * @param log
	 *            The given event log.
	 * @return A new transformed event log with added attributes and classifier.
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Eduardo Gonzalez Lopez de Murillas", email = "e.gonzalez@tue.nl")
	@Plugin(name = "Add Classifier based on Mapping of Activity Names", returnLabels = {
			"Transformed Log" }, returnTypes = { XLog.class }, parameterLabels = { "Log" }, userAccessible = true, categories = {
							PluginCategory.Enhancement }, help = "Adds a new classifier to the event log based on a user-configureable mapping from activity names of an existing classifier to the new classifier.")
	public XLog mapActivityNameDefault(UIPluginContext context, XLog log) {
		ActivityMappingParameters parameters = new ActivityMappingParameters();
		ActivityMappingDialog dialog = new ActivityMappingDialog(log, parameters);
		InteractionResult result = context.showWizard("Activity Mapping wizard", true, true, dialog);
		if (result != InteractionResult.FINISHED) {
			return null;
		}
		XLog newLog = (XLog) log.clone();
		return addActivityMappingName(context.getProgress(), newLog, parameters);
	}

	/**
	 * Transform using provided parameter values.
	 * 
	 * @param context
	 *            The given GUI-aware plug-in context.
	 * @param log
	 *            The given event log.
	 * @return A new transformed event log with added attributes and classifier.
	 */
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = "Eduardo Gonzalez Lopez de Murillas", email = "e.gonzalez@tue.nl")
	@Plugin(name = "Add Classifier based on Mapping of Activity Names (in place)", returnLabels = {}, returnTypes = {}, mostSignificantResult = -1, parameterLabels = {
			"Log" }, userAccessible = true, categories = {
					PluginCategory.Enhancement }, help = "Adds a new classifier to the event log based on a user-configureable mapping from activity names of an existing classifier to the new classifier. "
							+ "This variant does not copy the log, but modifies the existing log 'in place'.")
	public void mapActivityNameWithoutCopy(UIPluginContext context, XLog log) {
		ActivityMappingParameters parameters = new ActivityMappingParameters();
		ActivityMappingDialog dialog = new ActivityMappingDialog(log, parameters);
		InteractionResult result = context.showWizard("Activity Mapping wizard", true, true, dialog);
		if (result != InteractionResult.FINISHED) {
			return;
		}
		addActivityMappingName(context.getProgress(), log, parameters);
	}

	/**
	 * The actual transformation an event log given parameter values.
	 * 
	 * @param progress
	 * @param newLog
	 * @param parameters
	 * @return
	 */
	private XLog addActivityMappingName(Progress progress, XLog newLog, ActivityMappingParameters parameters) {

		String keyName = parameters.getNewClassifierName() + "-activityName";
		XEventAttributeClassifier newClassifier = new XEventAttributeClassifier(parameters.getNewClassifierName(),
				keyName);
		newLog.getClassifiers().add(newClassifier);

		XEventClassifier classifier = parameters.getClassifier();

		/*
		 * Inform the progress bar when we're done.
		 */
		progress.setMaximum(newLog.size());

		for (XTrace trace : newLog) {
			for (XEvent event : trace) {

				String identity = classifier.getClassIdentity(event);

				String newIdentity = parameters.getActivityMapping().get(identity);

				if (newIdentity == null) {
					newIdentity = identity;
				}

				event.getAttributes().put(keyName, new XAttributeLiteralImpl(keyName, newIdentity));
			}

			progress.inc();
		}

		/*
		 * Return the model.
		 */
		return newLog;
	}
}
