package org.processmining.logenhancement.view.comparison;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Event;
import org.processmining.framework.util.ui.widgets.traceview.ProMTraceView.Trace;
import org.processmining.log.utils.XUtils;
import org.processmining.plugins.utils.sequencedistance.GenericLevenshteinDistance;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance.DistanceResult;
import org.processmining.plugins.utils.sequencedistance.SequenceDistance.Equivalence;

public class LogComparisonPlugin {

	private final static class TraceEntryImpl implements TraceEntry {

		private XTrace traceA;
		private XTrace traceB;
		private int maxLength;
		private int minLength;
		private String name;
		private Trace<Event> combinedTrace;
		private int distance;
		private DistanceResult<XEvent> distanceMeasureResult;

		private TraceEntryImpl(String name) {
			super();
			this.name = name;
		}

		public String getName() {
			return name;
		}

		public int getMaxLength() {
			return maxLength;
		}

		public int getMinLength() {
			return minLength;
		}

		public void setTraceA(XTrace trace) {
			traceA = trace;
			maxLength = Math.max(maxLength, trace.size());
			minLength = Math.min(minLength, trace.size());
		}

		public void setTraceB(XTrace trace) {
			traceB = trace;
			maxLength = Math.max(maxLength, trace.size());
			minLength = Math.min(minLength, trace.size());
		}

		public XTrace getTraceB() {
			return traceB;
		}

		public XTrace getTraceA() {
			return traceA;
		}

		public void buildCombinedTrace(SequenceDistance<XEvent> distanceMeasure) {
			if (traceA != null && traceB != null) {

				distanceMeasureResult = distanceMeasure.computeAlignment(traceA, traceB, new Equivalence<XEvent>() {

					public boolean equals(XEvent a, XEvent b) {
						return XConceptExtension.instance().extractName(a)
								.equals(XConceptExtension.instance().extractName(b));
					}
				});

				distance = distanceMeasureResult.getDistance();

			} else {
				distanceMeasureResult = null;
				distance = Integer.MAX_VALUE;
			}
		}

		public Trace<? extends Event> getCombinedTrace() {
			return combinedTrace;
		}

		public int getDistance() {
			return distance;
		}

		public DistanceResult<XEvent> getDistanceResult() {
			return distanceMeasureResult;
		}

	}

	public interface TraceEntry {

		/**
		 * @return common concept:name of both traces 
		 */
		String getName();		

		XTrace getTraceA();
		XTrace getTraceB();

		/**
		 * @return minimum length of traces A, B
		 */
		int getMinLength();
		/**
		 * @return maximum length of traces A, B
		 */
		int getMaxLength();

		DistanceResult<XEvent> getDistanceResult();
		int getDistance();

		/**
		 * @return 
		 */
		Trace<? extends Event> getCombinedTrace();

	}

	public interface LogComparisonResult {

		Collection<TraceEntry> getTraceEntries();

	}

	@Plugin(name = "Compare Event Logs (Levenshtein Distance)", returnLabels = { "Log Comparison" }, returnTypes = { LogComparisonResult.class }, userAccessible = true, parameterLabels = {
			"Log A", "Log B" }, mostSignificantResult = 1)
	@UITopiaVariant(affiliation = UITopiaVariant.EHV, author = " F. Mannhardt", email = "f.mannhardt@tue.nl")
	public LogComparisonResult compareLogsPlugin(PluginContext context, XLog logA, XLog logB) throws Exception {
		context.getProgress().setCaption("Comparing logs ...");
		return compareLogs(logA, logB);
	}

	public LogComparisonResult compareLogs(XLog logA, XLog logB) {
		SequenceDistance<XEvent> distanceMeasure = new GenericLevenshteinDistance<>();
		return compareLogsUsing(logA, logB, distanceMeasure);
	}

	private LogComparisonResult compareLogsUsing(XLog logA, XLog logB, SequenceDistance<XEvent> distanceMeasure) {
		final Collection<TraceEntry> entries = buildTraceEntries(distanceMeasure, logA, logB);
		return new LogComparisonResult() {

			public Collection<TraceEntry> getTraceEntries() {
				return entries;
			}
		};
	}

	private Collection<TraceEntry> buildTraceEntries(SequenceDistance<XEvent> distanceMeasure, XLog logA, XLog logB) {

		Map<String, TraceEntry> entriesByName = new HashMap<>();

		for (XTrace trace : logA) {
			String name = XUtils.getConceptName(trace);
			if (name != null) {
				TraceEntryImpl traceEntry = new TraceEntryImpl(name);
				entriesByName.put(name, traceEntry);
				traceEntry.setTraceA(trace);
			} else {
				System.out.println("Trace is missing concept:name " + trace);
			}
		}

		for (XTrace trace : logB) {
			String name = XUtils.getConceptName(trace);
			if (name != null) {
				TraceEntryImpl traceEntry = (TraceEntryImpl) entriesByName.get(name);
				if (traceEntry == null) {
					traceEntry = new TraceEntryImpl(name);
				}
				traceEntry.setTraceB(trace);
			} else {
				System.out.println("Trace is missing concept:name " + trace);
			}
		}

		for (TraceEntry e : entriesByName.values()) {
			((TraceEntryImpl) e).buildCombinedTrace(distanceMeasure);
		}

		return entriesByName.values();
	}

}
